/*
COPYRIGHT NOTICE
Copyright (c) 2012  Scott C. Neu, Daniel J. Valentino, and Arthur W. Toga

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

import java.awt.BorderLayout;
import java.awt.Point;
import java.awt.color.ColorSpace;
import java.awt.event.MouseEvent;
import java.awt.image.BufferedImage;
import java.awt.image.DataBuffer;
import java.awt.image.DataBufferByte;
import java.io.File;
import java.io.IOException;
import java.util.Iterator;
import java.util.Vector;
import javax.imageio.ImageIO;
import javax.imageio.ImageReader;
import javax.imageio.stream.FileImageInputStream;
import javax.swing.JFrame;
import org.medtoolbox.jviewbox.BufferedImageUtilities;
import org.medtoolbox.jviewbox.DefaultImageRenderingHints;
import org.medtoolbox.jviewbox.ImageRenderingHints;
import org.medtoolbox.jviewbox.LookUpTable;
import org.medtoolbox.jviewbox.LinearLookUpTable;
import org.medtoolbox.jviewbox.imagesource.BufferedImageWrapper;
import org.medtoolbox.jviewbox.viewport.Viewport;
import org.medtoolbox.jviewbox.viewport.ViewportCluster;
import org.medtoolbox.jviewbox.viewport.ViewportGrid;
import org.medtoolbox.jviewbox.viewport.ViewportTool;
import org.medtoolbox.jviewbox.viewport.ViewportToolBar;

/**
 * Example that illustrates how to use the LONI Image I/O plugins to display
 * images with the LONI JViewBox package.
 * <p>
 * NOTE:  Running requires the plugin corresponding to the format of the
 *        file whose images are being displayed as well as the LONI JViewBox
 *        library jar file.
 *
 * @version 21 April 2005
 */
public class JViewBoxImageViewer
{
  /**
   * Constructs a JViewBox Image Viewer.
   *
   * @param fileName Name of the file to read an image from.
   */
  public JViewBoxImageViewer(String fileName)
    {
      try {

	// Test the file existence
	File file = new File(fileName);
	if ( !file.exists() || !file.isFile() || !file.canRead() ) {
	  System.out.println("Cannot read \"" + fileName + "\"");
	  System.exit(1);
	}

	// Find an Image Reader for the file
	ImageReader imageReader = _getImageReader(file);
	if (imageReader == null) {
	  System.out.println("Cannot decode \"" + fileName + "\"");
	  System.exit(1);
	}

	// Read all the images from the file
	Vector images = new Vector();
        for (int i = 0; i < imageReader.getNumImages(true); i++) {
	  images.add( imageReader.read(i) );
	}

	// Wrap each image in a Buffered Image Wrapper
	Vector wrappers = new Vector( images.size() );
	Iterator iter = images.iterator();
	while ( iter.hasNext() ) {
	  BufferedImage bi = (BufferedImage)iter.next();
	  wrappers.add( new BufferedImageWrapper(bi, _getHints(bi)) );
	}

	// Create a ToolBar and add Tools to it
	ViewportToolBar toolBar = new ViewportToolBar();
	toolBar.addTool(new PanTool(), "Pan");
	toolBar.addTool(new ZoomTool(), "Zoom");
	toolBar.addTool(new WindowLevelTool(), "Window/Level");
	toolBar.addTool(new ScrollTool(), "Scroll");

	// Create a Viewport Grid to display the images
	ViewportGrid grid = new ViewportGrid(wrappers);

	// Link the Tool Bar to the ViewportGrid
	grid.registerViewportToolBar(toolBar);

        // Create a JFrame for the images
        JFrame frame = new JFrame(fileName);
        frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);

	// Add the Tool Bar and Viewport Grid to this frame
	frame.getContentPane().add(toolBar, BorderLayout.NORTH);
	frame.getContentPane().add(grid, BorderLayout.CENTER);

        // Display the JFrame
        frame.setSize(800, 600);
        frame.setVisible(true);
      }

      catch (Exception e) {
	e.printStackTrace();
	System.exit(1);
      }
    }

  /**
   * Main method.
   *
   * @param args Command line parameters.
   */
  public static void main(String[] args) 
    {
      if (args.length < 1) {
	System.out.println("ARGS:  <file>");
	System.exit(1);
      }

      // Read images from the file and display using JViewBox
      new JViewBoxImageViewer(args[0]);
    }

  /**
   * Gets an Image Reader that can decode the specified file.
   *
   * @param file File to find an Image Reader for.
   *
   * @return Image Reader that can decode the specified file, or null if none
   *         was found.
   *
   * @throws IOException If an I/O error occurs.
   */
  private ImageReader _getImageReader(File file) throws IOException
    {
      Object input = new FileImageInputStream(file);
      Object origInput = input;

      // First look for a magic number in the header
      Iterator iter = ImageIO.getImageReaders(input);

      // No Image Reader is found
      if ( !iter.hasNext() ) {

	// Then look in the directory of the file (e.g., ANALYZE, AFNI)
	input = file;
	iter = ImageIO.getImageReaders(input);

	// No Image Reader is found
	if( !iter.hasNext() ) {
	  input = origInput;

	  // Then look at the file name suffix
	  String temp = file.getName();
	  String[] strings = temp.split("\\.");

	  if (strings.length > 1) {
	      int l = strings.length;
	      String suffix = strings[l-1];
	      iter = ImageIO.getImageReadersBySuffix(suffix);
	  }
	  if ( !iter.hasNext() ) {
	      if (strings.length > 2) {
		  int l = strings.length;
		  String suffix = strings[l-2] + "." + strings[l-1];
		  iter = ImageIO.getImageReadersBySuffix(suffix);
	      }
	  }

	  // No Image Reader found
	  if ( !iter.hasNext() ) { return null; }
	}
      }

      // Set the Input Stream of the first Image Reader returned
      ImageReader imageReader = (ImageReader)iter.next();
      imageReader.setInput(input);

      // Return the Image Reader
      return imageReader;
    }

  /**
   * Gets hints for rendering the Buffered Image.
   *
   * @param bi Buffered Image to render.
   *
   * @return Hints for rendering the Buffered Image.
   */
  private ImageRenderingHints _getHints(BufferedImage bi)
    {
      // Determine the number of bits per pixel
      int bitsPerPixel = 16;
      DataBuffer dataBuffer = bi.getRaster().getDataBuffer();
      if (dataBuffer instanceof DataBufferByte) { bitsPerPixel = 8; }

      // Create the hints
      LookUpTable lut = BufferedImageUtilities.createLUTFromProperties(bi);
      int[] bitsPerSample = new int[1];
      bitsPerSample[0] = bitsPerPixel;
      return new DefaultImageRenderingHints(new Integer(ColorSpace.TYPE_GRAY),
					    new Integer(1), new Boolean(true),
					    lut, bitsPerSample, bitsPerSample,
					    new Double(1), "Normal");
    }

  /**
   * Tool for panning an image inside a Viewport.  Copied from JViewBox demo.
   * <p>
   * When the mouse cursor is pressed and dragged in a Viewport, the image in 
   * the Viewport moves along with the cursor until the mouse button is
   * released.
   */
  private class PanTool extends ViewportTool
  {
    /** Anchor point relative to which the pan offset is calculated. */
    private final Point _anchor = new Point();

    /** Viewport being panned. */
    private Viewport _vp = null;

    /** Constructs a PanTool. */
    public PanTool()
    {
	super("Pan Tool", "Pan an Image", "", "");
    }

    /**
     * Invoked when a mouse button is pressed on a ViewportCluster.
     *
     * @param vpc ViewportCluster which received the event.
     * @param vp Viewport which the mouse cursor was on during the event;
     *		 <code>null</code> if the mouse was outside all visible
     *		 Viewports in the ViewportCluster.
     * @param e Event created by the button press.
     * @param button The button that was pressed.
     */
    public void mousePressed(ViewportCluster vpc, Viewport vp, MouseEvent e, 
			     int button)
    {
	_vp = vp;

	if (_vp != null) {
	    // Set the anchor
	    _anchor.x = e.getX() - _vp.getPanX();
	    _anchor.y = e.getY() - _vp.getPanY();

	    // Save the current state
	    _vp.saveCurrentState();

	}
    }

    /**
     * Invoked when a mouse button is pressed and then the mouse is dragged
     * on a ViewportCluster.
     *
     * @param vpc ViewportCluster which received the event.
     * @param vp Viewport which the mouse cursor was on during the event;
     *		 <code>null</code> if the mouse was outside all visible
     *		 Viewports in the ViewportCluster.
     * @param e Event created by the button drag.
     * @param button The button that was pressed and dragged.
     */
    public void mouseDragged(ViewportCluster vpc, Viewport vp, MouseEvent e, 
			     int button)
    {
	if (_vp != null && vpc != null) {

	    // Set the new pan
	    _vp.setPan(e.getX() - _anchor.x, e.getY() - _anchor.y);

	    vpc.repaint(_vp);

	}
    }
  }

  /**
   * Tool for scaling an image inside a Viewport.  Copied from JViewBox demo.
   * <p>
   * When the mouse cursor is pressed and dragged in a Viewport, its vertical
   * movement affects the scale factor applied to the image in the Viewport.
   * An upward drag decreases the scale factor (minifying, zooming out) while a
   * downward drag increases the scale factor (magnifying, zooming in).
   * <p>
   * When the mouse is dragged with the left button down, the normal
   * sensitivity applies. If the mouse is dragged with any other button down,
   * a 5x finer sensitivity applies (i.e., the same amount of vertical mouse
   * movement translates to one fifth of change in scale factors.)
   * <p>
   * Maximum and minimum limits apply to the scale factor. The limits can be
   * specified in the constructor.
   */
  private class ZoomTool extends ViewportTool
  {
    /**
     * Mouse movement in pixels that translates to a change of 1.0 to the
     * scale factors.
     */
    private int _sensitivity;

    /** Minimum allowed value for the Y scale factor. */
    private double _minimumScale;

    /** Maximum allowed value for the Y scale factor. */
    private double _maximumScale;

    /** Original X scale factor. */
    private double _origScaleX;

    /** Original Y scale factor. */
    private double _origScaleY;

    /**
     * Baseline Y coordinate relative to which the scale factors are
     * calculated.
     */
    private int _anchorY;

    /** Viewport being zoomed. */
    private Viewport _vp = null;

    /** Constructs a ZoomTool with default settings. */
    public ZoomTool()
    {
	this(200);
    }

    /**
     * Constructs a ZoomTool with the specified sensitivity and default
     * upper and lower bounds on the scale factor.
     */
    public ZoomTool(int sensitivity)
    {
	this(sensitivity, 0.01, 5.0);
    }

    /**
     * Constructs a ZoomTool with the specified sensitivity and upper/lower
     * bounds on the scale factor.
     *
     * @param sensitivity Mouse movement in pixels that translates to a change
     *			  of 1.0 to the scale factors.
     * @param minimumScale Minimum value of the Y scale factor allowed.
     * @param maximumScale Maximum value of the Y scale factor allowed.
     *
     * @throws IllegalArgumentException If any of the scale sensitivity,
     *         minimum scale, and maximum scale is invalid.
     */
    public ZoomTool(int sensitivity, double minimumScale,
		    double maximumScale)
    {
	super("Zoom Tool", "Zoom an Image", "", "");

	// Check bounds
	if (sensitivity <= 0 || minimumScale < 0.0001 ||
	    maximumScale < 0.0001) {
	    throw new IllegalArgumentException("ZoomTool: A scale sensitivity"+
					       " of " + sensitivity +
					       " and a minimum scale of " +
					       minimumScale + " and a maximum"+
					       " scale of " + maximumScale +
					       " are not allowed in the " +
					       "constructor.");
	}

	_sensitivity = sensitivity;
	_minimumScale = minimumScale;
	_maximumScale = maximumScale;
    }

    /**
     * Invoked when a mouse button is pressed on a ViewportCluster.
     *
     * @param vpc ViewportCluster which received the event.
     * @param vp Viewport which the mouse cursor was on during the event;
     *		 <code>null</code> if the mouse was outside all visible
     *		 Viewports in the ViewportCluster.
     * @param e Event created by the button press.
     * @param button The button that was pressed.
     */
    public void mousePressed(ViewportCluster vpc, Viewport vp, MouseEvent e, 
			     int button)
    {
	_vp = vp;

	if (_vp != null) {
	    // Set the anchor and current scale
	    _anchorY = e.getY();
	    _origScaleX = _vp.getScaleX();
	    _origScaleY = _vp.getScaleY();

	    // Save the current state
	    _vp.saveCurrentState();
	}
    }

    /**
     * Invoked when a mouse button is pressed and then the mouse is dragged
     * on a ViewportCluster.
     *
     * @param vpc ViewportCluster which received the event.
     * @param vp Viewport which the mouse cursor was on during the event.
     * @param e Event created by the button drag.
     * @param button The button that was pressed and dragged.
     */
    public void mouseDragged(ViewportCluster vpc, Viewport vp, MouseEvent e, 
			     int button)
    {
	if (_vp != null && vpc != null) {

	    // Compute the change in the Y scale factors
	    double deltaScaleY = (double)(e.getY() - _anchorY) /
		(button == LEFT_MOUSE_BUTTON ? _sensitivity : _sensitivity*5);

	    // Compute the new scale factors
	    double scaleY = _origScaleY + deltaScaleY;
	    if (scaleY < _minimumScale) {
		scaleY = _minimumScale;

	    }
	    else if (scaleY > _maximumScale) {
		scaleY = _maximumScale;

	    }
	    double scaleX = scaleY * _origScaleX / _origScaleY;

	    _vp.setScale(scaleX, scaleY);
	    vpc.repaint(_vp);

	}
    }
  }

  /**
   * Tool for changing the window and level values of the LinearLookUpTable
   * applied to the image inside a Viewport. This tool only works on Viewports
   * with LinearLookUpTables.  Copied from JViewBox demo.
   * <p>
   * When the mouse cursor is pressed and dragged in a Viewport, its horizontal
   * movement affects the level value and the vertical movement affects the
   * window value. A drag to the left decreases the level value while a drag to
   * the right increases it. An upward drag decreases the window value while a
   * downward drag increases it.
   * <p>
   * When the mouse is dragged with the left button down, the normal
   * sensitivity applies. If the mouse is dragged with any other button down,
   * a 5x finer sensitivity applies (i.e., the same amount of mouse movement
   * translates to one fifth of change in window and level values.)
   * <p>
   * If the mouse is double-clicked on a Viewport whose image is a 
   * <code>BufferedImage</code>, this tool calls
   * <code>BufferedImageUtilities.createLUTByScanning</code> to create a
   * look-up table to apply to the image.
   */
  private class WindowLevelTool extends ViewportTool
  {
    /**
     * Mouse movement in pixels that translates to the full scale of the 
     * LinearLookUpTable being adjusted.
     */
    private int _sensitivity;

    /** Original window value. */
    private int _origWindow;

    /** Original level value. */
    private int _origLevel;

    /** Anchor point relative to which window and level are calculated. */
    private final Point _anchor = new Point();

    /** Viewport whose LinearLookUpTable are being adjusted. */
    private Viewport _vp = null;

    /** LinearLookUpTable whose window/level values are being adjusted. */
    private LinearLookUpTable _vll = null;

    /**
     * Constructs a WindowLevelTool with a default sensitivity of 1000 pixels
     * to the full scale of a LinearLookUpTable.
     */
    public WindowLevelTool()
    {
	this(1000);
    }

    /**
     * Constructs a WindowLevelTool with the specified mouse window/level
     * sensitivity.
     *
     * @param sensitivity Mouse movement in pixels that translates
     *			  to the full scale of the LinearLookUpTable being
     *			  adjusted.
     *
     * @throws IllegalArgumentException If the window/level sensitivity
     *         is invalid.
     */
    public WindowLevelTool(int sensitivity)
    {
	super("Window/Level Tool", "Window/Level an Image", 
	      "Finer Window/Level", "Finer Window/Level");

	if (sensitivity <= 0) {
	    throw new IllegalArgumentException("WindowLevelTool: A window/" +
					       "level sensitivity of " +
					       sensitivity +
					       " is not allowed in the " +
					       "constructor.");
	}

	_sensitivity = sensitivity;
    }


    /**
     * Invoked when a mouse button is pressed on a ViewportCluster.
     *
     * @param vpc ViewportCluster which received the event.
     * @param vp Viewport which the mouse cursor was on during the event;
     *		 <code>null</code> if the mouse was outside all visible
     *		 Viewports in the ViewportCluster.
     * @param e Event created by the button press.
     * @param button The button that was pressed.
     */
    public void mousePressed(ViewportCluster vpc, Viewport vp, MouseEvent e, 
			     int button)
    {
	// Determine if there is a Linear LUT to window/level
	if (vp != null && vp.getLut() instanceof LinearLookUpTable) {
	    _vp = vp;

	    // Save the current state
	    vp.saveCurrentState();

	    // Set the anchor and original window/level values
	    _anchor.x = e.getX();
	    _anchor.y = e.getY();
	    _vll = (LinearLookUpTable)vp.getLut();
	    _origWindow = _vll.getWindow();
	    _origLevel = _vll.getLevel();

	}
	else {
	    _vp = null;
	    _vll = null;
	}
    }

    /**
     * Invoked when a mouse button is pressed and then the mouse is dragged
     * on a ViewportCluster.
     *
     * @param vpc ViewportCluster which received the event.
     * @param vp Viewport which the mouse cursor was on during the event.
     * @param e Event created by the button drag.
     * @param button The button that was pressed and dragged.
     */
    public void mouseDragged(ViewportCluster vpc, Viewport vp, MouseEvent e, 
			     int button)
    {
	if (_vll != null && _vp != null && vpc != null) {

	    // Find out the real sensitivity
	    double sensitivity = (double)_sensitivity / _vll.getSize();
	    if (button != LEFT_MOUSE_BUTTON) {
		sensitivity *= 5.0;
	    }

	    // Set the new window and level values
	    int window = (int)(_origWindow + 
			       (e.getY() - _anchor.y) / sensitivity);
	    int level = (int)(_origLevel + 
			      (e.getX() - _anchor.x) / sensitivity);
	    _vll.setWindowLevel(window, level);

	    vpc.repaint(_vp);

	}
    }

    /**
     * Invoked when a mouse button is clicked (pressed and released without
     * movement in between) on a ViewportCluster. Override this method to
     * implement tool functions triggered by the specific event.
     *
     * @param vpc ViewportCluster which received the event.
     * @param vp Viewport which the mouse cursor was on during the event;
     *		 <code>null</code> if the mouse was outside all visible
     *		 Viewports in the ViewportCluster.
     * @param e Event created by the button click.
     * @param button The button that was clicked.
     */
    public void mouseClicked(ViewportCluster vpc, Viewport vp, MouseEvent e, 
			     int button)
    {
	// Respond to double click only
	if (vp != null && e.getClickCount() == 2) {

	    // Can only handle BufferedImage
	    if (vp.getLut() instanceof LinearLookUpTable &&
		vp.getImage() instanceof BufferedImage) {

		// Save the current state
		vp.saveCurrentState();

		// Create a LUT by scanning the image for min/max
		BufferedImage image = (BufferedImage)vp.getImage();
		LookUpTable lut =
		    BufferedImageUtilities.createLUTByScanning(image);
		if (lut != null) {
		    vp.setLut(lut);
		    vpc.repaint(vp);
		}
	    }
	}
    }
  }

  /**
   * Tool for scrolling through the Viewports of a ViewportCluster.  Copied
   * from JViewBox demo.
   * <p>
   * When the mouse cursor is pressed and dragged downward on a
   * ViewportCluster, the ViewportCluster's offset is increased to bring
   * succeeding Viewports in the sequence into display. If the mouse is
   * dragged upward instead, the offset is decreased to bring preceding
   * Viewports in the sequence into display. If the mouse is simply clicked
   * (pressed and released without being dragged), the ViewportCluster is
   * scrolled forward by one unit on a left button click and backward by one
   * unit otherwise.
   */
  private class ScrollTool extends ViewportTool
  {
    /** Vertical mouse movement in pixels that translates to 1 scroll unit. */
    private int _scrollSensitivity;

    /** Baseline Y coordinate used to calculate the amount to scroll. */
    private int _anchorY;

    /** ViewportCluster being scrolled. */
    private ViewportCluster _vpc;

    /** Constructs a ScrollTool with default settings. */
    public ScrollTool()
    {
	this(20);
    }

    /**
     * Constructs a ScrollTool with the specified mouse scroll sensitivity.
     *
     * @param scrollSensitivity Vertical mouse movement in pixels that
     *				translates to 1 scroll unit.
     *
     * @throws IllegalArgumentException If the scroll sensitivity is invalid.
     */
    public ScrollTool(int scrollSensitivity)
    {
	super("Scroll Tool", "Scroll Through Images", "", "");

	// Check bounds
	if (scrollSensitivity <= 0) {
	    throw new IllegalArgumentException("ScrollTool: A scroll " +
					       "sensitivity of " +
					       scrollSensitivity + " is not " +
					       "allowed in the constructor.");
	}

	_scrollSensitivity = scrollSensitivity;
    }

    /**
     * Invoked when a mouse button is pressed on a ViewportCluster.
     *
     * @param vpc ViewportCluster which received the event.
     * @param vp Viewport which the mouse cursor was on during the event;
     *		 <code>null</code> if the mouse was outside all visible
     *		 Viewports in the ViewportCluster.
     * @param e Event created by the button press.
     * @param button The button that was pressed.
     */
    public void mousePressed(ViewportCluster vpc, Viewport vp, MouseEvent e, 
			     int button)
    {
	_vpc = null;

	// Set the anchor
	if (vpc != null) {
	    _vpc = vpc;
	    _anchorY = e.getY();
	}
    }

    /**
     * Invoked when a mouse button is pressed and then the mouse is dragged
     * on a ViewportCluster.
     *
     * @param vpc ViewportCluster which received the event.
     * @param vp Viewport which the mouse cursor was on during the event;
     *		 <code>null</code> if the mouse was outside all visible
     *		 Viewports in the ViewportCluster.
     * @param e Event created by the button drag.
     * @param button The button that was pressed and dragged.
     */
    public void mouseDragged(ViewportCluster vpc, Viewport vp, MouseEvent e, 
			     int button)
    {
	if (_vpc != null) {

	    // Determine how much to scroll
	    int scroll = (e.getY() - _anchorY) / _scrollSensitivity;

	    if (scroll != 0) {
		// Reset the anchor
		_anchorY = e.getY();

		// Scroll
		if (_vpc instanceof ViewportGrid) {
		    ((ViewportGrid)_vpc).gridScroll(scroll);
		}
		else {
		    _vpc.scroll(scroll);
		}

		// Repaint the ViewportCluster
		vpc.repaint();
	    }
	}
    }

    /**
     * Invoked when a mouse button is clicked (pressed and released without
     * movement in between) on a ViewportCluster.
     *
     * @param vpc ViewportCluster which received the event.
     * @param vp Viewport which the mouse cursor was on during the event;
     *		 <code>null</code> if the mouse was outside all visible
     *		 Viewports in the ViewportCluster.
     * @param e Event created by the button click.
     * @param button The button that was clicked.
     */
    public void mouseClicked(ViewportCluster vpc, Viewport vp, MouseEvent e, 
			     int button)
    {
	if (vpc != null) {
	    // Scroll forward for the left button, backward otherwise
	    int scroll = (button == LEFT_MOUSE_BUTTON) ? +1 : -1;

	    // Scroll
	    if (vpc instanceof ViewportGrid) {
		((ViewportGrid)vpc).gridScroll(scroll);
	    }
	    else {
		vpc.scroll(scroll);
	    }

	    // Repaint the ViewportCluster
	    vpc.repaint();
	}
    }
  }
}
