/*
  COPYRIGHT NOTICE
  Copyright (c) 2012  Scott C. Neu and Arthur W. Toga

  This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package edu.ucla.loni.freesurfer;

import java.util.Vector;

/**
 * Volume header for the MGH (Massachusetts General Hospital) file format.
 *
 * @version 19 March 2009
 */
public class MghVolumeHeader
{
    /** Version of the header. */
    private int _version;

    /** Width of each image (in voxels). */
    private int _width;

    /** Height of each image (in voxels). */
    private int _height;

    /** Number of images in a volume. */
    private int _depth;

    /** Number of scalar components per voxel. */
    private int _nframes;

    /** Data type of the image. */
    private int _type;

    /** Degrees of freedom. */
    private int _dof;

    /** Optional floats. */
    private Vector _floats;

    /** Optional tags. */
    private Vector _tags;

    /**
     * Constructs an MGH Volume Header.
     *
     * @param version Version of the header.
     * @param width Width of each image (in voxels).
     * @param height Height of each image (in voxels).
     * @param depth Number of images in a volume.
     * @param nframes Number of scalar components per voxel.
     * @param type Data type of the image.
     * @param dof Degrees of freedom.
     */
    public MghVolumeHeader(int version, int width, int height, int depth,
			   int nframes, int type, int dof)
    {
	_version = version;
	_width = width;
	_height = height;
	_depth = depth;
	_nframes = nframes;
	_type = type;
	_dof = dof;

	// Options
	_floats = new Vector();
	_tags = new Vector();
    }

    /**
     * Sets the float values.
     *
     * @param tr Repetition time (in milliseconds).
     */
    public void setFloats(float tr)
    {
	_floats = new Vector();

	_floats.add( new MghFloat(MghFloat.TR, tr) );
    }

    /**
     * Sets the float values.
     *
     * @param tr Repetition time (in milliseconds).
     * @param flipAngle Flip angle (in radians).
     */
    public void setFloats(float tr, float flipAngle)
    {
	setFloats(tr);

	_floats.add( new MghFloat(MghFloat.FLIP_ANGLE, flipAngle) );
    }

    /**
     * Sets the float values.
     *
     * @param tr Repetition time (in milliseconds).
     * @param flipAngle Flip angle (in radians).
     * @param te Echo time (in milliseconds).
     */
    public void setFloats(float tr, float flipAngle, float te)
    {
	setFloats(tr, flipAngle);

	_floats.add( new MghFloat(MghFloat.TE, te) );
    }

    /**
     * Sets the float values.
     *
     * @param tr Repetition time (in milliseconds).
     * @param flipAngle Flip angle (in radians).
     * @param te Echo time (in milliseconds).
     * @param ti Inversion time (in milliseconds).
     */
    public void setFloats(float tr, float flipAngle, float te, float ti)
    {
	setFloats(tr, flipAngle, te);

	_floats.add( new MghFloat(MghFloat.TI, ti) );
    }

    /**
     * Sets the float values.
     *
     * @param tr Repetition time (in milliseconds).
     * @param flipAngle Flip angle (in radians).
     * @param te Echo time (in milliseconds).
     * @param ti Inversion time (in milliseconds).
     * @param fov Field of view.
     */
    public void setFloats(float tr, float flipAngle, float te, float ti,
			  float fov)
    {
	setFloats(tr, flipAngle, te, ti);

	_floats.add( new MghFloat(MghFloat.FOV, fov) );
    }

    /**
     * Gets the float values.
     *
     * @return Float values.
     */
    public MghFloat[] getFloats()
    {
	return (MghFloat[])_floats.toArray(new MghFloat[0]);
    }

    /**
     * Adds a tag.
     *
     * @param code Tag code.
     * @param value Tag value.
     */
    public void addTag(int code, String value)
    {
	_tags.add( new MghTag(code, value) );
    }

    /**
     * Gets all the tags.
     *
     * @return All the tags.
     */
    public MghTag[] getTags()
    {
	return (MghTag[])_tags.toArray(new MghTag[0]);
    }

    /**
     * Gets the version of the header.
     *
     * @return Version of the header.
     */
    public int getVersion()
    {
	return _version;
    }

    /**
     * Gets the width of each image (in voxels).
     *
     * @return Width of each image (in voxels).
     */
    public int getWidth()
    {
	return _width;
    }

    /**
     * Gets the height of each image (in voxels).
     *
     * @return Height of each image (in voxels).
     */
    public int getHeight()
    {
	return _height;
    }

    /**
     * Gets the number of images in a volume.
     *
     * @return Number of images in a volume.
     */
    public int getDepth()
    {
	return _depth;
    }

    /**
     * Gets the number of scalar components per voxel.
     *
     * @return Number of scalar components per voxel.
     */
    public int getNframes()
    {
	return _nframes;
    }

    /**
     * Gets the data type of the image.
     *
     * @return Data type of the image.
     */
    public int getType()
    {
	return _type;
    }

    /**
     * Gets the degrees of freedom.
     *
     * @return Degrees of freedom.
     */
    public int getDof()
    {
	return _dof;
    }
}
