/*
  COPYRIGHT NOTICE
  Copyright (c) 2012  Scott C. Neu and Arthur W. Toga

  This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package edu.ucla.loni.freesurfer.plugin;

import edu.ucla.loni.blobio.BlobReader;
import edu.ucla.loni.freesurfer.CurvMeta;
import java.awt.image.DataBufferByte;
import java.io.IOException;
import javax.imageio.metadata.IIOMetadata;
import javax.imageio.stream.ImageInputStream;

/**
 * Blob Reader for the Free Surfer "New" Curv Format.
 *
 * @version 30 April 2012
 */
public class CurvBlobReader extends BlobReader
{
    /** Metadata decoded from the input source. */
    private CurvMeta _meta;

    /** Offset to the beginning of the blob data. */
    private long _beginBlobOffset;

    /** Offset to the end of the blob data. */
    private long _endBlobOffset;

    /**
     * Constructs a Curv Blob Reader.
     *
     * @param origProvider The Blob Reader SPI that instantiated this Object.
     */
    public CurvBlobReader(CurvBlobReaderSpi origProvider)
    {
	super(origProvider);
    }

    /**
     * Sets the input source to the specified Object.  The input source must be
     * set before any of the query or read methods are used.  If the input is
     * null, any currently set input source will be removed.
     *
     * @param input Object which is the source of the input.
     *
     * @throws IllegalArgumentException If the input is not an instance of one
     *                                  of the classes returned by the
     *                                  originating service provider's
     *                                  getInputTypes method.
     */
    public void setInput(Object input)
    {
	super.setInput(input);

	// Remove any decoded metadata
	_meta = null;
    }

    /**
     * Returns the number of blobs available from the current input source.
     *
     * @param maxSize Maximum size of a blob.  A blob larger than the maximum
     *                size will be subdivided into blobs smaller than this size.
     *
     * @return Number of blobs available from the current input source.
     *
     * @throws IOException If an error occurs reading the information from the
     *                     input source.
     * @throws IllegalStateException If the input source has not been set.
     */
    public int getNumberOfBlobs(int maxSize) throws IOException
    {
	// Check the size
	if (maxSize <= 0) { return 0; }

	// Determine the blob data size
	_getMeta();

	// Compute the number of blobs
	long blobSize = _endBlobOffset - _beginBlobOffset;
	return (int)Math.ceil(blobSize/(double)maxSize);
    }

    /**
     * Returns an IIOMetadata object representing the metadata associated with
     * the input source as a whole (i.e., not associated with any particular
     * blob), or null if the reader does not support reading metadata, is
     * set to ignore metadata, or if no metadata is available.
     *
     * @return An IIOMetadata object, or null.
     *
     * @throws IOException If an error occurs during reading.
     * @throws IllegalStateException If the input source has not been set.
     */
    public IIOMetadata getStreamMetadata() throws IOException
    {
	return new CurvMetadata( _getMeta() );
    }

    /**
     * Returns an IIOMetadata object containing metadata associated with the
     * given blob, or null if no metadata is available.
     *
     * @param blobIndex Index of the blob to retrieve.
     * @param maxSize Maximum size of a blob.  A blob larger than the maximum
     *                size will be subdivided into blobs smaller than this size.
     *
     * @return An IIOMetadata object, or null.
     *
     * @throws IOException If an error occurs during reading.
     * @throws IllegalStateException If the input source has not been set.
     * @throws IndexOutOfBoundsException If the supplied index is out of bounds.
     */
    public IIOMetadata getBlobMetadata(int blobIndex, int maxSize)
	throws IOException
    {
	// Not applicable
	return null;
    }

    /**
     * Returns the specified binary large object (BLOB).
     *
     * @param blobIndex Index of the blob to retrieve.
     * @param maxSize Maximum size of a blob.  A blob larger than the maximum
     *                size will be subdivided into blobs smaller than this size.
     *
     * @return Binary object.
     *
     * @throws IOException If an error occurs during reading.
     * @throws IllegalStateException If the input source has not been set.
     * @throws IndexOutOfBoundsException If the supplied index is out of bounds.
     */
    public DataBufferByte getBlob(int blobIndex, int maxSize) throws IOException
    {
	// Check the size
	if (maxSize <= 0) { return new DataBufferByte(0); }

	// Get the total number of blobs
	int numOfBlobs = getNumberOfBlobs(maxSize);

	// Check the blob index
	if (blobIndex < 0 || blobIndex >= numOfBlobs) {
	    String msg = "Blob index of " + blobIndex + " is not in the " +
		"range (0," + (numOfBlobs-1) + ")";
	    throw new IndexOutOfBoundsException(msg);
	}

	// Move to the beginning of the blob
	ImageInputStream imageStream = _getInputStream();
	imageStream.seek(_beginBlobOffset + blobIndex*maxSize);

	// Not last blob
	if (blobIndex < numOfBlobs-1) {
	    DataBufferByte blob = new DataBufferByte(maxSize);
	    imageStream.readFully( blob.getData() );
	    return blob;
	}

	// Last blob
	long blobSize = _endBlobOffset - imageStream.getStreamPosition();
	DataBufferByte blob = new DataBufferByte( (int)blobSize );
	imageStream.readFully( blob.getData() );
	return blob;
    }

    /**
     * Allows any resources held by this object to be released.  It is important
     * for applications to call this method when they know they will no longer
     * be using this Blob Reader.  Otherwise, the reader may continue to hold on
     * to resources indefinitely.
     */
    public void dispose()
    {
	// Attempt to close the input stream
	try { if (_input != null) { ((ImageInputStream)_input).close(); } }
	catch (Exception e) {}

	_input = null;
    }

    /**
     * Gets metadata from the input source.
     *
     * @return Metadata from the input source.
     *
     * @throws IOException If an I/O error occurs.
     * @throws IllegalStateException If the input has not been set.
     */
    private CurvMeta _getMeta() throws IOException
    {
	// Metadata has already been decoded
	if (_meta != null) { return _meta; }

	// Read the magic number
	ImageInputStream imageStream = _getInputStream();
	byte[] magicBytes = new byte[3];
	imageStream.readFully(magicBytes);

	// Read the number of vertices
	int vertexCount = imageStream.readInt();

	// Read the number of faces
	int faceCount = imageStream.readInt();

	// Read the number of values per vertex
	int valsPerVertex = imageStream.readInt();

	// Beginning of blob data
	_beginBlobOffset = imageStream.getStreamPosition();

	// Calculate the blob data size:
	//   1 vertex value = (float)
	long valueSize = 4*valsPerVertex*vertexCount;

	// End of blob data
	_endBlobOffset = _beginBlobOffset + valueSize;

	// Verify that the blob data exists
	if (_endBlobOffset > imageStream.length()) {
	    throw new IOException("Not enough data exists");
	}

	// Create the metadata
	_meta = new CurvMeta(vertexCount, faceCount, valsPerVertex);
	return _meta;
    }

    /**
     * Gets the input stream.
     *
     * @return Image Input Stream to read data from.
     *
     * @throws IllegalStateException If the input has not been set.
     */
    private ImageInputStream _getInputStream()
    {
	// No input has been set
	if (_input == null) {
	    String msg = "No input has been set.";
	    throw new IllegalStateException(msg);
	}

	// Return the Image Input Stream
	return (ImageInputStream)_input;
    }
}
