/*
  COPYRIGHT NOTICE
  Copyright (c) 2012  Scott C. Neu and Arthur W. Toga

  This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package edu.ucla.loni.ge;

import java.io.ByteArrayOutputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.util.Iterator;
import javax.imageio.stream.ImageOutputStream;

/**
 * Encoder for the header of the GE 5.0 format.
 *
 * @version 4 April 2011
 */
public class GEHeaderEncoder
{
    /** Output stream for the header. */
    private ImageOutputStream _imageStream;

    /**
     * Constructs a GE Header Encoder.
     *
     * @param imageStream Output stream for the header.
     */
    public GEHeaderEncoder(ImageOutputStream imageStream) 
    {
	_imageStream = imageStream;
    }

    /**
     * Writes the header.
     *
     * @param header Header to encode.
     *
     * @throws IOException If an I/O error occurs.
     */
    public void writeHeader(GEHeader header) throws IOException
    {
	// Define the subheader offsets and sizes
	int preHeaderSize = 124;
	int suiteInfoOffset = 2304;
	int suiteInfoSize = 114;
	int examInfoOffset = 2418;
	int examInfoSize = 1024;
	int seriesInfoOffset = 3442;
	int seriesInfoSize = 1020;
	int imageInfoOffset = 4462;
	int imageInfoSize = 1022;
	int pixelDataOffset = 7904;

	// Allocate memory for the header
	ByteArrayOutputStream outputStream =
	    new ByteArrayOutputStream(pixelDataOffset);

	// Write the preheader
	outputStream.write(_getPreHeaderBytes(header.getGEImageInfo(),
					      preHeaderSize, pixelDataOffset));

	// Write the subheader offsets and sizes (32 bytes)
	DataOutputStream dataStream = new DataOutputStream(outputStream);
	dataStream.writeInt(suiteInfoOffset);
	dataStream.writeInt(suiteInfoSize);
	dataStream.writeInt(examInfoOffset);
	dataStream.writeInt(examInfoSize);
	dataStream.writeInt(seriesInfoOffset);
	dataStream.writeInt(seriesInfoSize);
	dataStream.writeInt(imageInfoOffset);
	dataStream.writeInt(imageInfoSize);

	// Skip to the suite bytes
	byte[] zeroes = new byte[suiteInfoOffset - preHeaderSize - 32];
	outputStream.write(zeroes);

	// Write the GE Suite Info
	outputStream.write( _getSuiteInfoBytes(header.getGESuiteInfo(),
					       suiteInfoSize) );

	// Write the GE Exam Info
	outputStream.write( _getExamInfoBytes(header.getGEExamInfo(),
					      examInfoSize) );

	// Write the GE Series Info
	outputStream.write( _getSeriesInfoBytes(header.getGESeriesInfo(),
						seriesInfoSize) );

	// Write the GE Image Info
	outputStream.write( _getImageInfoBytes(header.getGEImageInfo(),
					       imageInfoSize) );

	// Skip to the pixel data
	zeroes = new byte[pixelDataOffset - imageInfoOffset - imageInfoSize];
	outputStream.write(zeroes);

	// Write the encoded bytes in memory
	byte[] encodedBytes = outputStream.toByteArray();
	byte[] bytes = new byte[pixelDataOffset];
	System.arraycopy(encodedBytes, 0, bytes, 0, encodedBytes.length);
	_imageStream.write(bytes);
    }

    /**
     * Gets the preheader encoded in a byte array.
     *
     * @param imageInfo GE Image Info to encode values from.
     * @param arraySize Size of the byte array to return.
     * @param pixelDataOffset Number of bytes from the start of the stream to
     *                        the pixel data.
     *
     * @return Byte array containing the encoded preheader.
     *
     * @throws IOException If an error occurs while encoding.
     */
    private byte[] _getPreHeaderBytes(GEImageInfo imageInfo, int arraySize,
				      int pixelDataOffset)
	throws IOException
    {
	// Allocate memory for the preheader
	ByteArrayOutputStream outStream = new ByteArrayOutputStream(arraySize);

	// Write the magic number
	int magicNumber = 0x494d4746;
	_writeNumber( new Integer(magicNumber), outStream );

	// Write the pixel data offset
	_writeNumber( new Integer(pixelDataOffset), outStream );

	// Write the image width and height
	String name = GEImageInfo.I_MATRIX_X;
	int width = ((Short)imageInfo.getElementValue(name)).intValue();
	name = GEImageInfo.I_MATRIX_Y;
	int height = ((Short)imageInfo.getElementValue(name)).intValue();

	_writeNumber( new Integer(width), outStream );
	_writeNumber( new Integer(height), outStream );

	// Write the number of bits per pixel
	name = GEImageInfo.SCREEN_FORMAT;
	int bitsPerPixel = ((Short)imageInfo.getElementValue(name)).intValue();
	_writeNumber( new Integer(bitsPerPixel), outStream );

	// Write the flag for no compression and rectangular
	_writeNumber( new Integer(1), outStream );

	// Copy the bytes to an array of the required size
	byte[] encodedBytes = outStream.toByteArray();
	byte[] bytes = new byte[arraySize];
	System.arraycopy(encodedBytes, 0, bytes, 0, encodedBytes.length);

	// Return the encoded bytes
	return bytes;
    }

    /**
     * Gets the Suite Info encoded in a byte array.
     *
     * @param suiteInfo GE Suite Info to encode.
     * @param arraySize Size of the byte array to return.
     *
     * @return Byte array containing the encoded Suite Info.
     *
     * @throws IOException If an error occurs while encoding.
     */
    private byte[] _getSuiteInfoBytes(GESuiteInfo suiteInfo, int arraySize)
	throws IOException
    {
	// Allocate memory for the Suite Info
	ByteArrayOutputStream outStream = new ByteArrayOutputStream(arraySize);

	// Write all the elements
	Iterator iter = suiteInfo.getElementNames();
	while ( iter.hasNext() ) {
	    String name = (String)iter.next();
	    Object value = suiteInfo.getElementValue(name);

	    // Element value is a String
	    if (value instanceof String) {
		int size = 0;
		if ( GESuiteInfo.SUITE_ID.equals(name) ) { size = 4; }
		if ( GESuiteInfo.PRODUCT_ID.equals(name) ) { size = 13; }
		if ( GESuiteInfo.VERSION_CR.equals(name) ) { size = 2; }
		if ( GESuiteInfo.VERSION_CU.equals(name) ) { size = 2; }

		// Write the string value
		_writeString( (String)value, size, outStream );
	    }

	    // Element value is a Character
	    else if (value instanceof Character) {
		_writeCharacter( (Character)value, outStream );
	    }

	    // Element value is a Number
	    else { _writeNumber( (Number)value, outStream ); }
	}

	// Copy the bytes to an array of the required size
	byte[] encodedBytes = outStream.toByteArray();
	byte[] bytes = new byte[arraySize];
	System.arraycopy(encodedBytes, 0, bytes, 0, encodedBytes.length);

	// Return the encoded bytes
	return bytes;
    }

    /**
     * Gets the Exam Info encoded in a byte array.
     *
     * @param examInfo GE Exam Info to encode.
     * @param arraySize Size of the byte array to return.
     *
     * @return Byte array containing the encoded Exam Info.
     *
     * @throws IOException If an error occurs while encoding.
     */
    private byte[] _getExamInfoBytes(GEExamInfo examInfo, int arraySize)
	throws IOException
    {
	// Allocate memory for the Exam Info
	ByteArrayOutputStream outStream = new ByteArrayOutputStream(arraySize);

	// Write all the elements
	Iterator iter = examInfo.getElementNames();
	while ( iter.hasNext() ) {
	    String name = (String)iter.next();
	    Object value = examInfo.getElementValue(name);

	    // Element value is a String
	    if (value instanceof String) {
		int size = 0;
		if ( GEExamInfo.SUITE_ID.equals(name) ) { size = 4; }
		if ( GEExamInfo.HOSPITAL_NAME.equals(name) ) { size = 32; }
		if ( GEExamInfo.PATIENT_ID.equals(name) ) { size = 12; }
		if ( GEExamInfo.PATIENT_NAME.equals(name) ) { size = 25; }
		if ( GEExamInfo.PATIENT_HISTORY.equals(name) ) { size = 61; }
		if ( GEExamInfo.REQ_NUMBER.equals(name) ) { size = 13; }
		if ( GEExamInfo.REF_PHYSICIAN.equals(name) ) { size = 33; }
		if ( GEExamInfo.DIAGRAD.equals(name) ) { size = 33; }
		if ( GEExamInfo.OPERATOR.equals(name) ) { size = 4; }
		if ( GEExamInfo.EXAM_DESCRIPTION.equals(name) ) { size = 23; }
		if ( GEExamInfo.EXAM_TYPE.equals(name) ) { size = 3; }
		if ( GEExamInfo.HOST.equals(name) ) { size = 9; }
		if ( GEExamInfo.ALLOC_KEY.equals(name) ) { size = 13; }
		if ( GEExamInfo.VERSION_CREATION.equals(name) ) { size = 2; }
		if ( GEExamInfo.VERSION_NOW.equals(name) ) { size = 2; }
		if ( GEExamInfo.SERIES_KEY.equals(name) ) { size = 4; }
		if ( GEExamInfo.UNST_SERIES_KEY.equals(name) ) { size = 4; }
		if ( GEExamInfo.UNARCH_SERIES_KEY.equals(name) ) { size = 4; }
		if ( GEExamInfo.PROS_SERIES_KEY.equals(name) ) { size = 4; }
		if ( GEExamInfo.MODEL_KEY.equals(name) ) { size = 4; }
		if ( GEExamInfo.SYSTEM_ID.equals(name) ) { size = 16; }
		if ( GEExamInfo.SERVICE_ID.equals(name) ) { size = 16; }

		// Write the string value
		_writeString( (String)value, size, outStream );
	    }

	    // Element value is a Character
	    else if (value instanceof Character) {
		_writeCharacter( (Character)value, outStream );
	    }

	    // Element value is a Number
	    else { _writeNumber( (Number)value, outStream ); }

	    // Skip bytes if necessary after the element
	    if ( name.equals(GEExamInfo.DISK_ID) ) {
		outStream.write(new byte[1]);
	    }
	    else if ( name.equals(GEExamInfo.HOSPITAL_NAME) ) {
		outStream.write(new byte[2]);
	    }
	    else if ( name.equals(GEExamInfo.PATIENT_ID) ) {
		outStream.write(new byte[1]);
	    }
	    else if ( name.equals(GEExamInfo.HOST) ) {
		outStream.write(new byte[1]);
	    }
	    else if ( name.equals(GEExamInfo.UPDATA_CNT) ) {
		outStream.write(new byte[1]);
	    }
	}

	// Copy the bytes to an array of the required size
	byte[] encodedBytes = outStream.toByteArray();
	byte[] bytes = new byte[arraySize];
	System.arraycopy(encodedBytes, 0, bytes, 0, encodedBytes.length);

	// Return the encoded bytes
	return bytes;
    }

    /**
     * Gets the Series Info encoded in a byte array.
     *
     * @param seriesInfo GE Series Info to encode.
     * @param arraySize Size of the byte array to return.
     *
     * @return Byte array containing the encoded Series Info.
     *
     * @throws IOException If an error occurs while encoding.
     */
    private byte[] _getSeriesInfoBytes(GESeriesInfo seriesInfo, int arraySize)
	throws IOException
    {
	// Allocate memory for the Series Info
	ByteArrayOutputStream outStream = new ByteArrayOutputStream(arraySize);

	// Write all the elements
	Iterator iter = seriesInfo.getElementNames();
	while ( iter.hasNext() ) {
	    String name = (String)iter.next();
	    Object value = seriesInfo.getElementValue(name);

	    // Element value is a String
	    if (value instanceof String) {
		int size = 0;
		if ( GESeriesInfo.SUITE_ID.equals(name) ) { size = 4; }
		if ( GESeriesInfo.SERIES_DESCRIPTION.equals(name) ) {
		    size = 30;
		}
		if ( GESeriesInfo.PRIMARY_RECIEVER_HOST.equals(name) ) {
		    size = 9;
		}
		if ( GESeriesInfo.ARCHIVER_HOST.equals(name) ) { size = 9; }
		if ( GESeriesInfo.ANATOM_REFERENCE.equals(name) ) {size = 3;}
		if ( GESeriesInfo.PROTOCOL_NAME.equals(name) ) { size = 25; }
		if ( GESeriesInfo.ALLOC_KEY.equals(name) ) { size = 13; }
		if ( GESeriesInfo.VERSION_CREATED.equals(name) ) { size = 2; }
		if ( GESeriesInfo.VERSION_NOW.equals(name) ) { size = 2; }
		if ( GESeriesInfo.IMAGE_KEY.equals(name) ) { size = 4; }
		if ( GESeriesInfo.UNST_IMAGE_KEY.equals(name) ) { size = 4; }
		if ( GESeriesInfo.UNARCH_IAMGE_KEY.equals(name) ) {size = 4;}

		// Write the string value
		_writeString( (String)value, size, outStream );
	    }

	    // Element value is a Character
	    else if (value instanceof Character) {
		_writeCharacter( (Character)value, outStream );
	    }

	    // Element value is a Number
	    else { _writeNumber( (Number)value, outStream ); }

	    // Skip bytes if necessary after the element
	    if ( name.equals(GESeriesInfo.DISK_ID) ) {
		outStream.write(new byte[1]);
	    }
	    else if ( name.equals(GESeriesInfo.ANATOM_REFERENCE) ) {
		outStream.write(new byte[1]);
	    }
	    else if ( name.equals(GESeriesInfo.PROTOCOL_NAME) ) {
		outStream.write(new byte[1]);
	    }
	    else if ( name.equals(GESeriesInfo.START_RAS) ) {
		outStream.write(new byte[1]);
	    }
	    else if ( name.equals(GESeriesInfo.END_RAS) ) {
		outStream.write(new byte[1]);
	    }
	    else if ( name.equals(GEMRImageInfo.ALLOC_KEY) ) {
		outStream.write(new byte[1]);
	    }
	}

	// Copy the bytes to an array of the required size
	byte[] encodedBytes = outStream.toByteArray();
	byte[] bytes = new byte[arraySize];
	System.arraycopy(encodedBytes, 0, bytes, 0, encodedBytes.length);

	// Return the encoded bytes
	return bytes;
    }

    /**
     * Gets the Image Info encoded in a byte array.
     *
     * @param imageInfo GE Image Info to encode.
     * @param arraySize Size of the byte array to return.
     *
     * @return Byte array containing the encoded Image Info.
     *
     * @throws IOException If an error occurs while encoding.
     */
    private byte[] _getImageInfoBytes(GEImageInfo imageInfo, int arraySize)
	throws IOException
    {
	// Allocate memory for the Image Info
	ByteArrayOutputStream outStream = new ByteArrayOutputStream(arraySize);

	// Write all the elements
	Iterator iter = imageInfo.getElementNames();
	while ( iter.hasNext() ) {
	    String name = (String)iter.next();
	    Object value = imageInfo.getElementValue(name);

	    // Element value is a String
	    if (value instanceof String) {
		int size = 0;
		if ( GEImageInfo.SUITE_ID.equals(name) ) { size = 4; }
		if ( GEImageInfo.PIXEL_DATA_ID.equals(name) ) { size = 14; }
		if ( GEImageInfo.CONTRAST_IV.equals(name) ) { size = 17; }
		if ( GEImageInfo.CONTRAST_ORAL.equals(name) ) { size = 17; }
		if ( GEImageInfo.FOREIGN_IMAGE_REV.equals(name) ) {size = 4;}
		if ( GEImageInfo.ALLOC_KEY.equals(name) ) { size = 13; }
		if ( GEImageInfo.VERSION_CREATION.equals(name) ) {size = 2;}
		if ( GEImageInfo.VERSION_NOW.equals(name) ) { size = 2; }
		if ( GEImageInfo.SLOP_STR1.equals(name) ) { size = 16; }
		if ( GEImageInfo.SLOP_STR2.equals(name) ) { size = 16; }

		if (imageInfo instanceof GEMRImageInfo) {
		    if ( GEMRImageInfo.PULSE_SEQ_NAME.equals(name) ) {
			size = 33;
		    }
		    if ( GEMRImageInfo.PSD_NAME.equals(name) ) { size = 13; }
		    if ( GEMRImageInfo.COIL_NAME.equals(name) ) { size = 17; }
		    if ( GEMRImageInfo.PROJECTION_NAME.equals(name) ) {
			size = 13;
		    }
		}

		if (imageInfo instanceof GECTImageInfo) {
		    if ( GECTImageInfo.QCAL_FILE.equals(name) ) { size = 13; }
		    if ( GECTImageInfo.CAL_MOD_FILE.equals(name) ) {
			size = 13;
		    }
		    if ( GECTImageInfo.SCOUT_ANATOM_REF.equals(name) ) {
			size = 3;
		    }
		}

		// Write the string value
		_writeString( (String)value, size, outStream );
	    }

	    // Element value is a Character
	    else if (value instanceof Character) {
		_writeCharacter( (Character)value, outStream );
	    }

	    // Element value is a Number
	    else { _writeNumber( (Number)value, outStream ); }

	    // Skip bytes if necessary after the element
	    if ( name.equals(GEImageInfo.DISK_ID) ) {
		outStream.write(new byte[1]);
	    }
	    else if ( name.equals(GEImageInfo.LOCATION_RAS) ) {
		outStream.write(new byte[1]);
	    }

	    if (imageInfo instanceof GEMRImageInfo) {
		if ( name.equals(GEMRImageInfo.PULSE_SEQ_NAME) ) {
		    outStream.write(new byte[1]);
		}
		else if ( name.equals(GEMRImageInfo.PSD_NAME) ) {
		    outStream.write(new byte[1]);
		}
		else if ( name.equals(GEMRImageInfo.COIL_NAME) ) {
		    outStream.write(new byte[1]);
		}
		else if ( name.equals(GEMRImageInfo.ALLOC_KEY) ) {
		    outStream.write(new byte[1]);
		}
	    }

	    if (imageInfo instanceof GECTImageInfo){
		if ( name.equals(GECTImageInfo.SCOUT_ANATOM_REF) ) {
		    outStream.write(new byte[1]);
		}
		else if ( name.equals(GEMRImageInfo.ALLOC_KEY) ) {
		    outStream.write(new byte[1]);
		}
	    }
	}

	// Copy the bytes to an array of the required size
	byte[] encodedBytes = outStream.toByteArray();
	byte[] bytes = new byte[arraySize];
	System.arraycopy(encodedBytes, 0, bytes, 0, encodedBytes.length);

	// Return the encoded bytes
	return bytes;
    }

    /**
     * Writes the Number to the output stream.
     *
     * @param number Number to write to the output stream.
     * @param outputStream Output Stream to write to.
     *
     * @throws IOException If an error occurs during the writing.
     */
    private void _writeNumber(Number number, OutputStream outputStream)
	throws IOException
    {
	DataOutputStream dataStream = new DataOutputStream(outputStream);

	// Byte
	if (number instanceof Byte) {
	    int value = ((Byte)number).byteValue() & 0xffffffff;
	    dataStream.writeByte(value);
	}

	// Short
	else if (number instanceof Short) {
	    int value = ((Short)number).shortValue() & 0xffffffff;
	    dataStream.writeShort(value);
	}

	// Integer
	else if (number instanceof Integer) {
	    dataStream.writeInt( ((Integer)number).intValue() );
	}

	// Float
	else if (number instanceof Float) {
	    dataStream.writeFloat( ((Float)number).floatValue() );
	}

	// Double
	else if (number instanceof Double) {
	    dataStream.writeDouble( ((Double)number).doubleValue() );
	}

	// Unknown Number
	else {
	    String msg = "Unable to write a Number of type \"" +
		number.getClass().getName() + "\".";
	    throw new IOException(msg);
	}
    }

    /**
     * Writes the Character to the output stream.
     *
     * @param character Character to write to the output stream.
     * @param outputStream Output Stream to write to.
     *
     * @throws IOException If an error occurs during the writing.
     */
    private void _writeCharacter(Character character, OutputStream outputStream)
	throws IOException
    {
	DataOutputStream dataStream = new DataOutputStream(outputStream);
	int value = character.charValue() & 0xffffffff;
	dataStream.writeByte(value);
    }

    /**
     * Writes the String to the output stream.
     *
     * @param string String to write to the output stream.
     * @param length Length of the string to actually write.
     * @param outputStream Output Stream to write to.
     *
     * @throws IOException If an error occurs during the writing.
     */
    private void _writeString(String string, int length,
			      OutputStream outputStream)
	throws IOException
    {
	// Create a byte buffer of the specified length
	byte[] buffer = new byte[length];

	// Copy the bytes of the String to the byte buffer
	byte[] stringBytes = string.getBytes();
	int min = Math.min( buffer.length, stringBytes.length );
	System.arraycopy(stringBytes, 0, buffer, 0, min);

	// Write the bytes
	outputStream.write(buffer);
    }
}
