/*
  COPYRIGHT NOTICE
  Copyright (c) 2012  Scott C. Neu and Arthur W. Toga

  This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package edu.ucla.loni.jdicom.decoder;

import java.io.IOException;
import java.util.Vector;
import javax.imageio.stream.ImageInputStream;

/**
 * Methods for Data Set I/O.
 *
 * @version 11 March 2011
 */
public class DataSetIO
{
    /** Stream to read Data Sets from. */
    private ImageInputStream _stream;

    /** Tags of Data Elements whose values are converted to Data Set Readers. */
    private Vector _dataElementTags;

    /**
     * Constructs a Data Set I/O.
     *
     * @param stream Stream to read Data Sets from.
     */
    public DataSetIO(ImageInputStream stream)
    {
	_stream = stream;
	_dataElementTags = new Vector();
    }

    /**
     * Sets the Data Element to have Data Set Reader values.
     *
     * @param dataElementTag Tag of the Data Element.
     */
    public void setReadersForValues(String dataElementTag)
    {
	if ( !hasReadersForValues(dataElementTag) ) {
	    _dataElementTags.add(dataElementTag);
	}
    }

    /**
     * Determines if the Data Element is to have Data Set Readers for values.
     *
     * @param dataElementTag Tag of the Data Element.
     *
     * @return True if the Data Element is to have Data Set Readers for values;
     *         false otherwise.
     */
    public boolean hasReadersForValues(String dataElementTag)
    {
	return _dataElementTags.contains(dataElementTag);
    }

    /**
     * Gets a Data Set Reader that reads the Data Set at the current stream
     * position.
     *
     * @param params Decoding parameters.
     * @param dataSetLength Length (in bytes) of the encoded DataSet.
     *
     * @throws IOException If an I/O error occurs.
     */
    public DataSetReader getDataSetReader(DecodeParams params,
					  int dataSetLength) throws IOException
    {
	return new DataSetReader(_stream, params, dataSetLength);
    }

    /**
     * Gets a Data Set Reader that reads the Data Set at the current stream
     * position.
     *
     * @param params Decoding parameters.
     * @param dataElementDelimiterTag 8 digit hexadecimal number marking the
     *                                end of the encoded DataSet.
     *
     * @throws IOException If an I/O error occurs.
     */
    public DataSetReader getDataSetReader(DecodeParams params,
					  String dataElementDelimiterTag)
	throws IOException
    {
	return new DataSetReader(_stream, params, dataElementDelimiterTag);
    }
}
