/*
  COPYRIGHT NOTICE
  Copyright (c) 2012  Scott C. Neu and Arthur W. Toga

  This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package edu.ucla.loni.jdicom.decoder;

import edu.ucla.loni.jdicom.DataElement;
import java.util.Enumeration;
import java.util.NoSuchElementException;

/**
 * Parameters used to decode DICOM Data Elements.
 *
 * @version 11 March 2011
 */
public class DecodeParams
{
    /**
     * Little Endian byte ordering.
     * <p>
     * <b> PS 3.5-1998, BIG ENDIAN VERSUS LITTLE ENDIAN BYTE ORDERING
     *     [Section 7.3], Page 28: </b>
     * <em>
     * <p>
     * Little Endian byte ordering is defined as follows:
     * <p> <ul>
     * <li> In a binary number consisting of multiple bytes (e.g. a 32-bit
     *      unsigned integer value, the Group Number, the Element Number,
     *      etc.), the least significant byte shall be encoded first; with the
     *      remaining bytes encoded in increasing order of significance.
     * <li> In a character string consisting of multiple 8-bit single byte
     *      codes, the characters will be encoded in the order of occurrence
     *      in the string (left to right).
     * </ul>
     * </em>
     */
    public static final int LITTLE_ENDIAN = 0;

    /**
     * Big Endian byte ordering.
     * <p>
     * <b> PS 3.5-1998, BIG ENDIAN VERSUS LITTLE ENDIAN BYTE ORDERING
     *     [Section 7.3], Page 28: </b>
     * <em>
     * <p>
     * Big Endian byte ordering is defined as follows:
     * <p> <ul>
     * <li> In a binary number consisting of multiple bytes, the most
     *      significant byte shall be encoded first; with the remaining bytes
     *      encoded in decreasing order of significance.
     * <li> In a character string consisting of multiple 8-bit single byte
     *      codes, the characters will be encoded in the order of occurrence
     *      in the string (left to right).
     * </ul> 
     * </em>
     */
    public static final int BIG_ENDIAN = 1;

    /**
     * Implicit VR structure.
     * <p>
     * <b> PS 3.5-1998, DATA ELEMENT STRUCTURE WITH IMPLICIT VR
     *     [Section 7.1.3], Page 27: </b>
     * <em> 
     * When using the Implicit VR structure the Data Element shall be
     * constructed of three consecutive fields:  Data Element Tag, Value
     * Length, and Value.
     * </em>
     */
    public static final int IMPLICIT_VR = 0;

    /**
     * Explicit VR structure.
     * <p>
     * <b> PS 3.5-1998, DATA ELEMENT STRUCTURE WITH EXPLICIT VR
     *     [Section 7.1.2], Page 26: </b>
     * <em> 
     * When using the Explicit VR structures, the Data Element shall be
     * constructed of four consecutive fields:  Data Element Tag, VR, Value
     * Length, and Value.
     * </em>
     */
    public static final int EXPLICIT_VR = 1;

    /** Byte ordering. */
    private int _byteOrdering;

    /** Data element structure. */
    private int _structure;

    /** Pixel representation. */
    private int _pixelRepresentation;

    /** Manufacturer. */
    private String _manufacturer;

    /** Maximum amount of available memory. */
    private long _maxMemory;

    /** Constructs a Decode Params. */
    public DecodeParams()
    {
	_maxMemory = Runtime.getRuntime().maxMemory();
	_pixelRepresentation = -1;

	// Default to the Transfer Syntax defined for all Group 2 Data Elements
	//
	// PS 3.10-1996, DICOM FILE META INFORMATION
	// [Section 7.1], Page 14:
	// Except for the 128 byte preamble and the 4 byte prefix, the File
	// Meta Information shall be encoded using the Explicit VR Little
	// Endian Transfer Syntax (UID=1.2.840.10008.1.2.1) as defined in
	// DICOM PS 3.5.
	_byteOrdering = LITTLE_ENDIAN;
	_structure = EXPLICIT_VR;
    }

    /**
     * Gets the byte ordering.
     *
     * @return Byte ordering.
     */
    public int getByteOrdering()
    {
	return _byteOrdering;
    }

    /**
     * Gets the data element structure.
     *
     * @return Data element structure.
     */
    public int getStructure()
    {
	return _structure;
    }

    /**
     * Gets the pixel representation.
     *
     * @return Pixel representation or -1 if it is not known.
     */
    public int getPixelRepresentation()
    {
	return _pixelRepresentation;
    }

    /**
     * Gets the manufacturer.
     *
     * @return Manufacturer or null if it is not known.
     */
    public String getManufacturer()
    {
	return _manufacturer;
    }

    /**
     * Gets the maximum amount of available memory.
     *
     * @return Maximum amount of available memory.
     */
    public long getMaxMemory()
    {
	return _maxMemory;
    }

    /**
     * Updates the decode parameters after the Data Element is decoded.
     *
     * @param dataElement Decoded Data Element.
     *
     * @throws NoSuchElementException If a valueless Transfer Syntax is
     *                                specified.
     * @throws IllegalArgumentException If an unknown Transfer Syntax is found.
     */
    public void update(DataElement dataElement)
    {
	String tag = dataElement.getDataElementTag();

	// Transfer syntax
	if ( "00020010".equals(tag) ) {
	    _updateTransferSyntax(dataElement);
	    return;
	}

	// Pixel representation
	if ( "00280103".equals(tag) ) {
	    _updatePixelRepresentation(dataElement);
	    return;
	}

	// Manufacturer
	if ( "00080070".equals(tag) ) {
	    _updateManufacturer(dataElement);
	    return;
	}
    }

    /**
     * Updates the transfer syntax.
     *
     * @param dataElement Transfer syntax Data Element.
     *
     * @throws NoSuchElementException If a valueless Transfer Syntax is
     *                                specified.
     * @throws IllegalArgumentException If an unknown Transfer Syntax is found.
     */
    private void _updateTransferSyntax(DataElement dataElement)
    {
	// Look for the transfer syntax value
	Enumeration enun = dataElement.getValues();

	// Check that there is a value
	if ( !enun.hasMoreElements() ) {
	    String msg = "No value for the Transfer Syntax is found.";
	    throw new NoSuchElementException(msg);
	}
	String transferSyntax = (String)enun.nextElement();

	// Transfer Syntax value implies Little Endian byte ordering and
	// Explicit VR structure
	//
	// PS 3.5-1998, DICOM LITTLE ENDIAN TRANSFER SYNTAX (EXPLICIT VR)
	// [Section A.2], Page 43:
	// This DICOM Explicit VR Little Endian Transfer Syntax shall be
	// identified by a UID of Value "1.2.840.10008.1.2.1."
	//
	// PS 3.5-1998, TRANSFER SYNTAXES FOR ENCAPSULATION OF ENCODED PIXEL
	//              DATA
	// [Section A.4], Page 44:
	// a) The Data Elements contained in the Data Set structure shall be
	// encoded with Explicit VR (with a VR Field) as specified in
	// Section 7.1.2.
	// b) The encoding of the overall Data Set structure (Data Element Tags,
	// Value Length, etc.) shall be in Little Endian as specified in
	// Section 7.3.
	//
	// PS 3.5-1998, JPEG IMAGE COMPRESSION
	// [Section A.4.1], Page 47:
	// A DICOM Transfer Syntax for JPEG Image Compression shall be
	// identified by a UID value, appropriate to its JPEG coding process,
	// chosen from Table A.4-3.
	//
	// PS 3.5-1998, RLE COMPRESSION
	// [Section A.4.2], Page 48:
	// This transfer Syntax is identified by the UID value
	// 1.2.840.10008.1.2.5.
	if ( transferSyntax.equals("1.2.840.10008.1.2.1") ||
	     transferSyntax.startsWith("1.2.840.10008.1.2.4") ||
	     transferSyntax.startsWith("1.2.840.10008.1.2.5") )
	    {
		_byteOrdering = LITTLE_ENDIAN;
		_structure = EXPLICIT_VR;
		return;
	    }

	// Transfer Syntax value implies Big Endian byte ordering and
	// Explicit VR structure
	//
	// PS 3.5-1998, DICOM BIG ENDIAN TRANSFER SYNTAX (EXPLICIT VR)
	// [Section A.3], Page 44:
	// This DICOM Explicit VR Big Endian Transfer Syntax shall be identified
	// by a UID of Value "1.2.840.10008.1.2.2."
	if ( transferSyntax.equals("1.2.840.10008.1.2.2") ) {
	    _byteOrdering = BIG_ENDIAN;
	    _structure = EXPLICIT_VR;
	    return;
	}

	// Transfer Syntax value implies Little Endian byte ordering and
	// Implicit VR structure
	//
	// PS 3.5-1998, DICOM IMPLICIT VR LITTLE ENDIAN TRANSFER SYNTAX
	// [Section A.1], Page 42:
	// This DICOM Implicit VR Little Endian Transfer Syntax shall be
	// identified by a UID of Value "1.2.840.10008.1.2."
	if ( transferSyntax.equals("1.2.840.10008.1.2") ) {
	    _byteOrdering = LITTLE_ENDIAN;
	    _structure = IMPLICIT_VR;
	    return;
	}

	// This option is included to handle the remaining case, which is a
	// violation of the DICOM standard, yet is still encountered
	if ( transferSyntax.equals("BIG_ENDIAN_IMPLICIT_VR") ) {
	    _byteOrdering = BIG_ENDIAN;
	    _structure = IMPLICIT_VR;
	    return;
	}

	// Transfer Syntax value is unrecognized
	throw new IllegalArgumentException("Unknown Transfer Syntax " +
					   transferSyntax + " found.");
    }

    /**
     * Updates the pixel representation.
     *
     * @param dataElement Pixel representation Data Element.
     */
    private void _updatePixelRepresentation(DataElement dataElement)
    {
	// Look for the pixel representation
	Enumeration enun = dataElement.getValues();

	// Check that there is a value
	if ( !enun.hasMoreElements() ) { return; }

	// Store the value
	_pixelRepresentation = ((Integer)enun.nextElement()).intValue();
    }

    /**
     * Updates the manufacturer.
     *
     * @param dataElement Manufacturer Data Element.
     */
    private void _updateManufacturer(DataElement dataElement)
    {
	// Look for the manufacturer
	Enumeration enun = dataElement.getValues();

	// Check that there is a value
	if ( !enun.hasMoreElements() ) { return; }

	// Store the value
	_manufacturer = (String)enun.nextElement();
    }
}
