/*
  COPYRIGHT NOTICE
  Copyright (c) 2012  Scott C. Neu and Arthur W. Toga

  This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package edu.ucla.loni.jdicom.plugin;

import edu.ucla.loni.imageio.LoniImageWriteParam;
import edu.ucla.loni.jdicom.DataSet;
import edu.ucla.loni.jdicom.plugin.write.DicomWriter;
import java.awt.image.BufferedImage;
import java.awt.image.DataBuffer;
import java.awt.image.DataBufferByte;
import java.awt.image.DataBufferInt;
import java.awt.image.DataBufferUShort;
import java.awt.image.Raster;
import java.awt.image.RenderedImage;
import java.io.IOException;
import java.util.Enumeration;
import javax.imageio.IIOImage;
import javax.imageio.ImageTypeSpecifier;
import javax.imageio.ImageWriteParam;
import javax.imageio.ImageWriter;
import javax.imageio.metadata.IIOMetadata;
import javax.imageio.stream.ImageOutputStream;
import javax.imageio.spi.ImageWriterSpi;
import org.w3c.dom.Node;

/**
 * Image Writer for parsing and encoding part 10 DICOM images.
 *
 * @version 17 March 2011
 */
public class DicomImageWriter extends ImageWriter
{
    /** Data Set of the image sequence currently being written. */
    private DataSet _dataSet;

    /** DICOM Writer for the image sequence currently being written. */
    private DicomWriter _writer;

    /** Number of images in the sequence currently being written. */
    private int _numberOfImagesWritten;

    /**
     * Constructs a Dicom Image Writer.
     *
     * @param originatingProvider The ImageWriterSpi that instantiated this
     *                            Object.
     *
     * @throws IllegalArgumentException If the originating provider is not a
     *                                  DICOM Image Writer Spi.
     */
    public DicomImageWriter(ImageWriterSpi originatingProvider)
    {
	super(originatingProvider);

	// Originating provider must be a DICOM Image Writer Spi
	if ( !(originatingProvider instanceof DicomImageWriterSpi) ) {
	    String msg = "Originating provider must a DICOM Image Writer SPI.";
	    throw new IllegalArgumentException(msg);
	}
    }

    /**
     * Returns an IIOMetadata object containing default values for encoding a
     * stream of images.
     *
     * @param param An ImageWriteParam that will be used to encode the image, or
     *              null.
     *
     * @return An IIOMetadata object.
     */
    public IIOMetadata getDefaultStreamMetadata(ImageWriteParam param)
    {
	// Return a blank stream metadata object
	String name = DicomStreamMetadataFormat.NATIVE_METADATA_FORMAT_NAME;
	String className = DicomStreamMetadataFormat.class.getName();
	return new DicomMetadata(name, className, null);
    }

    /**
     * Returns an IIOMetadata object containing default values for encoding an
     * image of the given type.
     *
     * @param imageType An ImageTypeSpecifier indicating the format of the image
     *                  to be written later.
     * @param param An ImageWriteParam that will be used to encode the image, or
     *              null.
     *
     * @return An IIOMetadata object.
     */
    public IIOMetadata getDefaultImageMetadata(ImageTypeSpecifier imageType,
					       ImageWriteParam param)
    {
	// Return a blank image metadata object
	String name = DicomImageMetadataFormat.NATIVE_METADATA_FORMAT_NAME;
	String className = DicomImageMetadataFormat.class.getName();
	return new DicomMetadata(name, className, null);
    }

    /**
     * Returns an IIOMetadata object that may be used for encoding and
     * optionally modified using its document interfaces or other interfaces
     * specific to the writer plug-in that will be used for encoding.
     *
     * @param inData An IIOMetadata object representing stream metadata, used to
     *               initialize the state of the returned object.
     * @param param An ImageWriteParam that will be used to encode the image, or
     *              null.
     *
     * @return An IIOMetadata object, or null if the plug-in does not provide
     *         metadata encoding capabilities.
     *
     * @throws IllegalArgumentException If inData is null.
     */
    public IIOMetadata convertStreamMetadata(IIOMetadata inData,
					     ImageWriteParam param)
    {
	// Only recognize Dicom Metadata
	if (inData instanceof DicomMetadata) { return inData; }

	// Otherwise perform no conversion
	return null;
    }

    /**
     * Returns an IIOMetadata object that may be used for encoding and
     * optionally modified using its document interfaces or other interfaces
     * specific to the writer plug-in that will be used for encoding.
     *
     * @param inData An IIOMetadata object representing image metadata, used to
     *               initialize the state of the returned object.
     * @param imageType An ImageTypeSpecifier indicating the layout and color
     *                  information of the image with which the metadata will be
     *                  associated.
     * @param param An ImageWriteParam that will be used to encode the image, or
     *              null.
     *
     * @return An IIOMetadata object, or null if the plug-in does not provide
     *         metadata encoding capabilities.
     *
     * @throws IllegalArgumentException If either of inData or imageType is
     *                                  null.
     */
    public IIOMetadata convertImageMetadata(IIOMetadata inData,
					    ImageTypeSpecifier imageType,
					    ImageWriteParam param)
    {
	// Only recognize Dicom Metadata
	if (inData instanceof DicomMetadata) { return inData; }

	// Otherwise perform no conversion
	return null;
    }

    /**
     * Returns true if the methods that take an IIOImage parameter are capable
     * of dealing with a Raster (as opposed to RenderedImage) source image.
     *
     * @return True if Raster sources are supported.
     */
    public boolean canWriteRasters()
    {
	return true;
    }

    /**
     * Appends a complete image stream containing a single image and associated
     * stream and image metadata and thumbnails to the output.
     *
     * @param streamMetadata An IIOMetadata object representing stream metadata,
     *                       or null to use default values.
     * @param image An IIOImage object containing an image, thumbnails, and
     *              metadata to be written.
     * @param param An ImageWriteParam, or null to use a default
     *              ImageWriteParam.
     *
     * @throws IOException If an error occurs during writing.
     * @throws IllegalArgumentException If the image is null, if the stream
     *                                  metadata cannot be converted, or if the
     *                                  image type is unrecognized.
     * @throws IllegalStateException If the output has not been set.
     */
    public void write(IIOMetadata streamMetadata, IIOImage image,
		      ImageWriteParam param) throws IOException
    {
	// Write a sequence with 1 image
	prepareWriteSequence(streamMetadata);
	writeToSequence(image, param);
	endWriteSequence();
    }

    /**
     * Returns true if the writer is able to append an image to an image
     * stream that already contains header information and possibly prior
     * images.
     *
     * @return True If images may be appended sequentially.
     */
    public boolean canWriteSequence()
    {
	return true;
    }

    /**
     * Prepares a stream to accept a series of subsequent writeToSequence calls,
     * using the provided stream metadata object.  The metadata will be written
     * to the stream if it should precede the image data.
     *
     * @param streamMetadata A stream metadata object.
     *
     * @throws IOException If an error occurs during writing.
     * @throws IllegalArgumentException If the stream metadata is null or cannot
     *                                  be converted.
     * @throws IllegalStateException If the output has not been set.
     */
    public void prepareWriteSequence(IIOMetadata streamMetadata)
	throws IOException
    {
	// Convert the metadata to DICOM Metadata
	IIOMetadata metadata = convertStreamMetadata(streamMetadata, null);

	// Unable to convert the metadata
	if (metadata == null) {
	    String msg = "Unable to convert the stream metadata for encoding.";
	    throw new IllegalArgumentException(msg);
	}

	// Construct a Data Set from the metadata
	try {
	    String name = DicomStreamMetadataFormat.NATIVE_METADATA_FORMAT_NAME;
	    Node tree = metadata.getAsTree(name);
	    _dataSet = DicomMetadataConversions.toDataSet(tree);
	}
      
	catch (Exception e) {
	    String msg="Unable to construct a Data Set from the metadata tree.";
	    IllegalArgumentException excep = new IllegalArgumentException(msg);
	    excep.initCause(e);
	    throw excep;
	}

	// Create a new DICOM Writer for the image sequence
	String tSyntax = _getStringValue("00020010", _dataSet);
	_writer = new DicomWriter(_getOutputStream(), tSyntax);

	// Write the sequence header
	_writer.writeSequenceHeader(_dataSet);

	_numberOfImagesWritten = 0;
    }

    /**
     * Appends a single image and possibly associated metadata and thumbnails,
     * to the output.  The supplied ImageReadParam is ignored.
     *
     * @param image An IIOImage object containing an image, thumbnails, and
     *              metadata to be written.
     * @param param An ImageWriteParam or null to use a default ImageWriteParam.
     *
     * @throws IOException If an error occurs during writing.
     * @throws IllegalArgumentException If the image is null or if the image
     *                                  type is not recognized.
     * @throws IllegalStateException If the output has not been set, or
     *                               prepareWriteSequence has not been called.
     */
    public void writeToSequence(IIOImage image, ImageWriteParam param)
	throws IOException
    {
	// Check for the Data Set
	if (_dataSet == null) {
	    String msg = "The image sequence has not been prepared.";
	    throw new IllegalStateException(msg);
	}

	// Check for a null image
	if (image == null) {
	    String msg = "Cannot write a null image.";
	    throw new IllegalArgumentException(msg);
	}

	// Default write mode
	boolean writeMetadata = false;
	boolean writeImageData = true;

	// Check for the LONI write mode
	if (param instanceof LoniImageWriteParam) {
	    LoniImageWriteParam loniParam = (LoniImageWriteParam)param;
	    if ( loniParam.writeOnlyMetadata() ) {
		writeMetadata = true;
		writeImageData = false;
	    }
	    else if ( loniParam.writeImageAndMetadata() ) {
		writeMetadata = true;
		writeImageData = true;
	    }
	}

	// Convert the image metadata to a data set
	DataSet dataSet = null;
	IIOMetadata metadata = image.getMetadata();
	if (metadata != null) {
	    String name = DicomImageMetadataFormat.NATIVE_METADATA_FORMAT_NAME;
	    try {
		Node tree = metadata.getAsTree(name);
		dataSet = DicomMetadataConversions.toDataSet(tree);

		// If image metadata is supplied, use it
		writeMetadata = true;
	    }
	    catch (Exception e) {
		String msg = "Unable to make a data set from image metadata";
		IllegalArgumentException ie = new IllegalArgumentException(msg);
		ie.initCause(e);
		throw ie;
	    }
	}

	// Get a Data Buffer from the image
	Raster raster = image.getRaster();
	if (raster == null) {
	    RenderedImage renderedImage = image.getRenderedImage();

	    // If the Rendered Image is a Buffered Image, get it Raster directly
	    if (renderedImage instanceof BufferedImage) {
		raster = ((BufferedImage)renderedImage).getRaster();
	    }

	    // Otherwise get a copy of the Raster from the Rendered Image
	    else { raster = renderedImage.getData(); }
	}
	DataBuffer dataBuffer = raster.getDataBuffer();

	// Update the Listeners
	processImageStarted(_numberOfImagesWritten);

	// Get the image width and height
	int imageWidth = _getIntValue("00280011", _dataSet);
	int imageHeight = _getIntValue("00280010", _dataSet);

	// Write the image
	if (writeMetadata && !writeImageData) {
	    _writer.writeImageMetadata(dataSet);
	}
	else if (!writeMetadata && writeImageData) {
	    _writer.writeImage(dataBuffer, imageWidth, imageHeight);
	}
	else {
	    _writer.writeImage(dataSet, dataBuffer, imageWidth, imageHeight);
	}

	// Update the Listeners
	_numberOfImagesWritten++;
	if ( abortRequested() ) { processWriteAborted(); }
	else { processImageComplete(); }
    }

    /**
     * Completes the writing of a sequence of images begun with
     * prepareWriteSequence.  Any stream metadata that should come at the end of
     * the sequence of images is written out, and any header information at the
     * beginning of the sequence is patched up if necessary.
     *
     * @throws IOException If an error occurs during writing.
     * @throws IllegalStateException If the output has not been set, or
     *                               prepareWriteSequence has not been called.
     */
    public void endWriteSequence() throws IOException
    {
	// Check for the Data Set
	if (_dataSet == null) {
	    String msg = "The image sequence has not been prepared.";
	    throw new IllegalStateException(msg);
	}

	// Write the sequence footer
	_writer.writeSequenceFooter(_dataSet);

	// Clear
	_dataSet = null;
	_writer = null;
    }

    /**
     * Allows any resources held by this object to be released.  It is important
     * for applications to call this method when they know they will no longer
     * be using this ImageWriter.  Otherwise, the writer may continue to hold on
     * to resources indefinitely.
     */
    public void dispose()
    {
	// Attempt to close the output stream
	try { if (output != null) { ((ImageOutputStream)output).close(); } }
	catch (Exception e) {}

	output = null;
	_writer = null;
    }

    /**
     * Gets the string value of the specified Data Element Tag.
     *
     * @param dataElementTag Tag of the Data Element to get the value for.
     * @param dataSet Data Set to get the Data Element value from.
     *
     * @return String value of the Data Element, or "" if no value exists.
     */
    private String _getStringValue(String dataElementTag, DataSet dataSet)
    {
	try {
	    Enumeration enun =
		dataSet.getDataElement(dataElementTag).getValues();
	    String value = enun.nextElement().toString();

	    // Return value without beginning & ending spaces (and no leading +)
	    value = value.trim();
	    if ( value.charAt(0) == '+') { value = value.substring(1); }
	    return value;
	}

	// Return a blank string if no string value exists
	catch (Exception e) { return ""; }
    }

    /**
     * Gets the integer value of the specified Data Element Tag.
     *
     * @param dataElementTag Tag of the Data Element to get the value for.
     * @param dataSet Data Set to get the Data Element value from.
     *
     * @return Integral value of the Data Element, or -1 if no value exists.
     */
    private int _getIntValue(String dataElementTag, DataSet dataSet)
    {
	// Get the value as a String
	String value = _getStringValue(dataElementTag, dataSet);

	// Return the value as an integer
	try { return Integer.parseInt(value); }
	catch (Exception e) { return -1; }
    }

    /**
     * Gets the output stream.
     *
     * @return Image Output Stream to write to.
     *
     * @throws IllegalStateException If the output has not been set.
     */
    private ImageOutputStream _getOutputStream()
    {
	// No output has been set
	if (output == null) {
	    String msg = "No output has been set.";
	    throw new IllegalStateException(msg);
	}

	// Return the Image Output Stream
	return (ImageOutputStream)output;
    }
}
