/*
  COPYRIGHT NOTICE
  Copyright (c) 2012  Scott C. Neu and Arthur W. Toga

  This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package edu.ucla.loni.jdicom.plugin.node;

import edu.ucla.loni.Conversions;
import org.w3c.dom.Attr;
import org.w3c.dom.DOMException;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.TypeInfo;
import org.w3c.dom.UserDataHandler;

/**
 * DICOM Node that holds one attribute value.
 *
 * @version 24 February 2011
 */
public class AttributeNode extends DicomNode implements Attr, NamedNodeMap
{
    /** Name of the attribute. */
    private String _name;

    /** Value of the attribute. */
    private Object _value;

    /**
     * Constructs an Attribute Node.
     *
     * @param name Name of the attribute.
     * @param value Value of the attribute
     */
    public AttributeNode(String name, Object value)
    {
	_name = name;

	// If the value is a byte array, convert it to a Hex String
	if (value instanceof byte[]) {
	    value = Conversions.toHexString( (byte[])value );
	}

	// Convert nulls to the empty string
	else if (value == null) { value = ""; }

	_value = value;
    }

    /**
     * Gets the Node name.
     *
     * @return Name of the Node.
     */
    public String getNodeName()
    {
	return _name;
    }

    /**
     * Gets the Node value.
     *
     * @return Value of the Node.
     */
    public String getNodeValue() throws DOMException
    {
	if (_value instanceof String) { return (String)_value; }
	return _value.toString();
    }

    /**
     * Sets the Node value.
     *
     * @param value Value of the Node.
     */
    public void setNodeValue(String nodeValue) throws DOMException
    {
	setValue(nodeValue);
    }

    /**
     * Gets the Node type.
     *
     * @return Node type.
     */
    public short getNodeType()
    {
	return ATTRIBUTE_NODE;
    }

    /**
     * Gets a duplicate of this Node.  The duplicate Node has no parent.  If a
     * shallow clone is being performed, the duplicate Node will not have any
     * children or siblings.  If a deep clone is performed, the duplicate Node
     * will form the root of a completely cloned subtree.
     *
     * @param deep True for a recursive clone, false for cloning only the Node
     *             itself.
     *
     * @return Duplicate Node.
     */
    public Node cloneNode(boolean deep)
    {
	return new AttributeNode(_name, _value);
    }

    /**
     * Gets the name of this attribute.
     *
     * @return Name of the attribute.
     */
    public String getName()
    {
	return _name;
    }

    /**
     * Determines if this attribute is explicitly given a value in the original
     * document.
     *
     * @return True if this attribute is explicitly given a value; false o/w.
     */
    public boolean getSpecified()
    {
	return true;
    }

    /**
     * Gets the value of the attribute.
     *
     * @return Value of the attribute.
     */
    public String getValue()
    {
	if (_value instanceof String) { return (String)_value; }
	return _value.toString();
    }

    /**
     * Sets the value of the attribute.
     *
     * @param value Value for the attribute.
     */
    public void setValue(String value) throws DOMException
    {
	if (value == null) { value = ""; }
	_value = value;
    }

    /**
     * Gets the owner Element.
     *
     * @return Owner Element.
     */
    public Element getOwnerElement()
    {
	return null;
    }

    /**
     * Gets the named Node.
     *
     * @param name Name of the Node.
     *
     * @return Named Node or null if no such Node exists.
     */
    public Node getNamedItem(String name)
    {
	if ( _name.equals(name) ) { return this; }
	return null;
    }

    /**
     * Adds the named Node.
     *
     * @param arg Node to add.
     *
     * @return Replaced Node if it exists, otherwise null.
     *
     * @throws DOMException If no modification is allowed.
     */
    public Node setNamedItem(Node arg) throws DOMException
    {
	String msg = "No modifications are allowed";
	throw new DOMException(DOMException.NO_MODIFICATION_ALLOWED_ERR, msg);
    }

    /**
     * Removes the named Node.
     *
     * @param name Name of the Node to remove.
     *
     * @return Removed Node if it exists.
     *
     * @throws DOMException If no modification is allowed.
     */
    public Node removeNamedItem(String name) throws DOMException
    {
	String msg = "No modifications are allowed";
	throw new DOMException(DOMException.NO_MODIFICATION_ALLOWED_ERR, msg);
    }

    /**
     * Gets the indexed child Node.
     *
     * @param index Index of the child.
     *
     * @return Indexed child Node or null if the index is invalid.
     */
    public Node item(int index)
    {
	if (index == 0) { return this; }
	return null;
    }

    /**
     * Gets the number of children of this Node.
     *
     * @return Number of children of this Node.
     */
    public int getLength()
    {
	return 1;
    }

    /**
     * Gets the named Node specified by the local name and namespace URI.
     *
     * @param name Name of the Node.
     *
     * @return Named Node or null if no such Node exists.
     */
    public Node getNamedItemNS(String namespaceURI, String localName)
    {
	return getNamedItem(localName);
    }

    /**
     * Adds the named Node using its local name and namespace URI.
     *
     * @param arg Node to add.
     *
     * @return Replaced Node if it exists, otherwise null.
     *
     * @throws DOMException If no modification is allowed.
     */
    public Node setNamedItemNS(Node arg) throws DOMException
    {
	return setNamedItem(arg);
    }

    /**
     * Removes the named Node specified by the local name and namespace URI.
     *
     * @param name Name of the Node to remove.
     *
     * @return Removed Node if it exists.
     *
     * @throws DOMException If no modification is allowed.
     */
    public Node removeNamedItemNS(String namespaceURI, String localName)
	throws DOMException
    {
	return removeNamedItem(localName);
    }

    // Methods added to compile with Java 1.4.2 are below

    public boolean isId()
    {
        throw new DOMException(DOMException.NOT_SUPPORTED_ERR,
			       "Method not supported");
    }
}
