/*
  COPYRIGHT NOTICE
  Copyright (c) 2012  Scott C. Neu and Arthur W. Toga

  This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package edu.ucla.loni.jdicom.plugin.node;

import edu.ucla.loni.jdicom.DataElement;
import edu.ucla.loni.jdicom.DataSet;
import edu.ucla.loni.jdicom.plugin.DicomMetadataFormat;
import java.util.ArrayList;
import java.util.Enumeration;
import org.w3c.dom.DOMException;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;

/**
 * DICOM List Node that represents a Data Element.
 *
 * @version 6 April 2010
 */
public class DataElementNode extends DicomListNode
{
    /** Name of the Data Element. */
    private String _name;

    /** Value representation. */
    private AttributeNode _vr;

    /**
     * Constructs a Data Element Node.
     *
     * @param dataElement Data Element to represent.
     */
    public DataElementNode(DataElement dataElement)
    {
	this(dataElement.getDataElementTag(),
	     dataElement.getValueRepresentation(),
	     _getChildren(dataElement));
    }

    /**
     * Constructs a Data Element Node.
     *
     * @param name Name of the Data Element.
     * @param vr Value representation of the Data Element.
     */
    public DataElementNode(String name, String vr)
    {
	this(name, vr, new ArrayList(0));
    }

    /**
     * Constructs a Data Element Node.
     *
     * @param name Name of the Data Element.
     * @param vr Value representation of the Data Element.
     * @param children Child Nodes for the Data Element.
     */
    private DataElementNode(String name, String vr, ArrayList children)
    {
	super(children);

	_name = name;
	_vr = new AttributeNode(DicomMetadataFormat.VR_NAME, vr);
    }

    /**
     * Gets the tag of the Data Element.
     *
     * @return Tag of the Data Element.
     */
    public String getTag()
    {
	return _name;
    }

    /**
     * Gets the value representation of the Data Element.
     *
     * @return Value representation of the Data Element.
     */
    public String getVr()
    {
	return _vr.getValue();
    }

    /**
     * Gets the nested Data Sets of the Data Element.
     *
     * @return Nested Data Sets of the Data Element.
     */
    public DataSetNode[] getDataSets()
    {
	ArrayList sets = new ArrayList(getLength());

	// Find all the nested Data Sets
	for (int i = 0; i < getLength(); i++) {
	    Node child = item(i);
	    if (child instanceof DataSetNode) { sets.add(child); }
	}

	// Return the nested Data Sets
	return (DataSetNode[])sets.toArray(new DataSetNode[0]);
    }

    /**
     * Gets the first value of the Data Element.
     *
     * @return First value of the Data Element or null if no value exists.
     */
    public String getValue()
    {
	// Find the first Value Node
	for (int i = 0; i < getLength(); i++) {
	    Node child = item(i);
	    if (child instanceof ValueNode) {
		return ((ValueNode)child).getValue();
	    }
	}

	// No values
	return null;
    }

    /**
     * Gets all the values of the Data Element.
     *
     * @return All the values of the Data Element.
     */
    public String[] getValues()
    {
	ArrayList values = new ArrayList(getLength());

	// Find all the values
	for (int i = 0; i < getLength(); i++) {
	    Node child = item(i);
	    if (child instanceof ValueNode) {
		values.add( ((ValueNode)child).getValue() );
	    }
	}

	// Return the values
	return (String[])values.toArray(new String[0]);
    }

    /**
     * Adds a value to the Data Element.
     *
     * @param value Value to add.
     */
    public void addValue(Object value)
    {
	appendChild( new ValueNode(value) );
    }

    /**
     * Replaces all the values of the Data Element.
     *
     * @param value New value of the Data Element.
     */
    public void replaceValues(Object value)
    {
	// Remove all the current children
	removeAllChildren();

	// Add the value
	addValue(value);
    }

    /**
     * Replaces all the values of the Data Element.
     *
     * @param values New values of the Data Element.
     */
    public void replaceValues(Object[] values)
    {
	// Remove all the current children
	removeAllChildren();

	// Add the values
	for (int i = 0; i < values.length; i++) { addValue(values[i]); }
    }

    /**
     * Gets the Node name.
     *
     * @return Name of the Node.
     */
    public String getNodeName()
    {
	return _name;
    }

    /**
     * Gets the attributes of this Node.
     *
     * @return Attributes of this Node, or null if there are no attributes.
     */
    public NamedNodeMap getAttributes()
    {
	return _vr;
    }

    /**
     * Inserts the Node before the existing child Node.  If the existing child
     * Node is null, the new Node is inserted at the end of the list of
     * children.
     *
     * @param newChild Node to insert.
     * @param refChild Node before which the new Node is inserted.
     *
     * @return Node that is inserted.
     *
     * @throws DOMException If the child is not a Value or Data Set Node.
     * @throws DOMException If the reference child is not a child of this Node.
     */
    public Node insertBefore(Node newChild, Node refChild) throws DOMException
    {
	// Illegal new child
	if ( !(newChild instanceof ValueNode) &&
	     !(newChild instanceof DataSetNode) )
	    {
		String msg = "Illegal new child node";
		throw new DOMException(DOMException.WRONG_DOCUMENT_ERR, msg);
	    }

	return super.insertBefore(newChild, refChild);
    }

    /**
     * Replaces the child Node with the given child Node.
     *
     * @param newChild Node to insert.
     * @param oldChild Node being replaced.
     *
     * @return Node that is replaced.
     *
     * @throws DOMException If the child is not a Value or Data Set Node.
     * @throws DOMException If the old child is not a child of this Node.
     */
    public Node replaceChild(Node newChild, Node oldChild) throws DOMException
    {
	// Illegal new child
	if ( !(newChild instanceof ValueNode) &&
	     !(newChild instanceof DataSetNode) )
	    {
		String msg = "Illegal new child node";
		throw new DOMException(DOMException.WRONG_DOCUMENT_ERR, msg);
	    }

	return super.replaceChild(newChild, oldChild);
    }

    /**
     * Adds the Node to the end of the list of children Nodes.
     *
     * @param newChild Node to add.
     *
     * @return Node that is added.
     *
     * @throws DOMException If the child is not a Value or Data Set Node.
     */
    public Node appendChild(Node newChild) throws DOMException
    {
	// Illegal new child
	if ( !(newChild instanceof ValueNode) &&
	     !(newChild instanceof DataSetNode) )
	    {
		String msg = "Illegal new child node";
		throw new DOMException(DOMException.WRONG_DOCUMENT_ERR, msg);
	    }

	return super.appendChild(newChild);
    }

    /**
     * Gets a duplicate of this Node.  The duplicate Node has no parent.  If a
     * shallow clone is being performed, the duplicate Node will not have any
     * children or siblings.  If a deep clone is performed, the duplicate Node
     * will form the root of a completely cloned subtree.
     *
     * @param deep True for a recursive clone, false for cloning only the Node
     *             itself.
     *
     * @return Duplicate Node.
     */
    public Node cloneNode(boolean deep)
    {
	ArrayList children = new ArrayList(0);
	if (deep) {
	    children = new ArrayList( getLength() );

	    for (int i = 0; i < getLength(); i++) {
		Node child = item(i);
		children.add( child.cloneNode(true) );
	    }
	}

	return new DataElementNode(_name, _vr.getValue(), children);
    }

    /**
     * Determines if this Node has attributes.
     *
     * @return True if this Node has attributes; false otherwise.
     */
    public boolean hasAttributes()
    {
	return true;
    }

    /**
     * Creates child Nodes from the Data Element.
     *
     * @param dataElement Data Element to represent.
     *
     * @return Child Nodes from the Data Element.
     */
    private static ArrayList _getChildren(DataElement dataElement)
    {
	// Initialize the list
	ArrayList children = new ArrayList(dataElement.getNumberOfValues() +
					   dataElement.getNumberOfDataSets());

	// Add the values
	Enumeration enumValues = dataElement.getValues();
	while ( enumValues.hasMoreElements() ) {
	    Object value = enumValues.nextElement();
	    children.add( new ValueNode(value) );
	}

	// Add the Data Sets
	Enumeration enumSets = dataElement.getDataSets();
	while ( enumSets.hasMoreElements() ) {
	    DataSet nestedDataSet = (DataSet)enumSets.nextElement();
	    children.add( new DataSetNode(nestedDataSet) );
	}

	return children;
    }
}
