/*
  COPYRIGHT NOTICE
  Copyright (c) 2012  Scott C. Neu and Arthur W. Toga

  This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package edu.ucla.loni.jdicom.plugin.node;

import org.w3c.dom.Attr;
import org.w3c.dom.DOMException;
import org.w3c.dom.Document;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.TypeInfo;
import org.w3c.dom.UserDataHandler;

/**
 * Node for DICOM information.
 *
 * @version 24 February 2011
 */
public abstract class DicomNode implements Node
{
    /** Parent of this Node, or null if there is no parent. */
    private Node _parent;

    /** Node immediately preceding this Node or null if none exists. */
    private Node _previousSibling;

    /** Node immediately following this Node or null if none exists. */
    private Node _nextSibling;

    /** Constructs a DICOM Node. */
    protected DicomNode()
    {
    }

    /**
     * Gets the Node value.
     *
     * @return Value of the Node.
     */
    public String getNodeValue() throws DOMException
    {
	return null;
    }

    /**
     * Sets the Node value.
     *
     * @param value Value of the Node.
     *
     * @throws DOMException If no modification is allowed.
     */
    public void setNodeValue(String nodeValue) throws DOMException
    {
	String msg = "No modifications are allowed";
	throw new DOMException(DOMException.NO_MODIFICATION_ALLOWED_ERR, msg);
    }

    /**
     * Gets the parent of this Node.
     *
     * @return Parent of this Node, or null if there is no parent.
     */
    public Node getParentNode()
    {
	return _parent;
    }

    /**
     * Gets a Node List that contains all children of this Node.
     *
     * @return Node List that contains all children of this Node.
     */
    public NodeList getChildNodes()
    {
	return new NodeList() {
		public Node item(int index) { return null; }
		public int getLength() { return 0; }
	    };
    }

    /**
     * Gets the first child of this Node.
     *
     * @return First child of this Node or null if none exists.
     */
    public Node getFirstChild()
    {
	return null;
    }

    /**
     * Gets the last child of this Node.
     *
     * @return Last child of this Node or null if none exists.
     */
    public Node getLastChild()
    {
	return null;
    }

    /**
     * Gets the Node immediately preceding this Node.
     *
     * @return Node immediately preceding this Node or null if none exists.
     */
    public Node getPreviousSibling()
    {
	return _previousSibling;
    }

    /**
     * Gets the Node immediately following this Node.
     *
     * @return Node immediately following this Node or null if none exists.
     */
    public Node getNextSibling()
    {
	return _nextSibling;
    }

    /**
     * Gets the attributes of this Node.
     *
     * @return Attributes of this Node, or null if there are no attributes.
     */
    public NamedNodeMap getAttributes()
    {
	return null;
    }

    /**
     * Gets the Document associated with this node.
     *
     * @return Document associated with this node or null if none exists.
     */
    public Document getOwnerDocument()
    {
	return null;
    }

    /**
     * Inserts the Node before the existing child Node.  If the existing child
     * Node is null, the new Node is inserted at the end of the list of
     * children.  If the new child is already in the list of children, it is
     * first removed.
     *
     * @param newChild Node to insert.
     * @param refChild Node before which the new Node is inserted.
     *
     * @return Node that is inserted.
     *
     * @throws DOMException If the child is not a Data Element Node.
     * @throws DOMException If the reference child is not a child of this Node.
     */
    public Node insertBefore(Node newChild, Node refChild) throws DOMException
    {
	String msg = "No children allowed";
	throw new DOMException(DOMException.NO_MODIFICATION_ALLOWED_ERR, msg);
    }

    /**
     * Replaces the child Node with the given child Node.  If the new child is
     * already in the list of children, it is first removed.
     *
     * @param newChild Node to insert.
     * @param oldChild Node being replaced.
     *
     * @return Node that is replaced.
     *
     * @throws DOMException If the child is not a Data Element Node.
     * @throws DOMException If the old child is not a child of this Node.
     */
    public Node replaceChild(Node newChild, Node oldChild) throws DOMException
    {
	String msg = "No children allowed";
	throw new DOMException(DOMException.NO_MODIFICATION_ALLOWED_ERR, msg);
    }

    /**
     * Removes the child Node from the list of children Nodes.
     *
     * @param oldChild Node being removed.
     *
     * @return Node that is removed.
     *
     * @throws DOMException If the old child is not a child of this Node.
     */
    public Node removeChild(Node oldChild) throws DOMException
    {
	String msg = "No children allowed";
	throw new DOMException(DOMException.NO_MODIFICATION_ALLOWED_ERR, msg);
    }

    /**
     * Adds the Node to the end of the list of children Nodes.  If the Node
     * is already in the list, it is first removed.
     *
     * @param newChild Node to add.
     *
     * @return Node that is added.
     *
     * @throws DOMException If the child is not a Data Element Node.
     */
    public Node appendChild(Node newChild) throws DOMException
    {
	String msg = "No children allowed";
	throw new DOMException(DOMException.NO_MODIFICATION_ALLOWED_ERR, msg);
    }

    /**
     * Determines if this Node has child Nodes.
     *
     * @return True if this Node has child Nodes; false otherwise.
     */
    public boolean hasChildNodes()
    {
	return false;
    }

    /** Normalizes text Nodes. */
    public void normalize()
    {
    }

    /**
     * Determines if the feature is supported in the version.
     *
     * @param feature Name of the feature.
     * @param version Version number.
     *
     * @return True if the feature is supported in the version; false o/w.
     */
    public boolean isSupported(String feature, String version)
    {
	return false;
    }

    /**
     * Gets the namespace URI of this Node.
     *
     * @return Namespace URI of this Node.
     */
    public String getNamespaceURI()
    {
	return null;
    }

    /**
     * Gets the namespace prefix of this Node.
     *
     * @return Namespace prefix of this Node.
     */
    public String getPrefix()
    {
	return null;
    }

    /**
     * Sets the namespace prefix of this Node.
     *
     * @param prefix Namespace prefix of this Node.
     */
    public void setPrefix(String prefix) throws DOMException
    {
    }

    /**
     * Gets the local part of the qualified name of this Node.
     *
     * @return Local part of the qualified name of this Node.
     */
    public String getLocalName()
    {
	return getNodeName();
    }

    /**
     * Determines if this Node has attributes.
     *
     * @return True if this Node has attributes; false otherwise.
     */
    public boolean hasAttributes()
    {
	return false;
    }

    /**
     * Sets the parent of this Node.
     *
     * @param parent Parent of this Node, or null if there is no parent.
     */
    protected void _setParent(Node parent)
    {
	_parent = parent;
    }

    /**
     * Sets the Node immediately preceding this Node.
     *
     * @param previousSibling Node immediately preceding this Node or null if
     *                        none exists.
     */
    protected void _setPreviousSibling(Node previousSibling)
    {
	_previousSibling = previousSibling;
    }

    /**
     * Sets the Node immediately following this Node.
     *
     * @param nextSibling Node immediately following this Node or null if none
     *                    exists.
     */
    protected void _setNextSibling(Node nextSibling)
    {
	_nextSibling = nextSibling;
    }

    // Methods added to compile with Java 1.4.2 are below

    public void setIdAttribute(String name, boolean isId) throws DOMException
    {
        throw new DOMException(DOMException.NOT_SUPPORTED_ERR,
			       "Method not supported");
    }

    public void setIdAttributeNS(String namespaceURI, String localName,
                                 boolean isId) throws DOMException
    {
        throw new DOMException(DOMException.NOT_SUPPORTED_ERR,
			       "Method not supported");
    }

    public void setIdAttributeNode(Attr idAttr, boolean isId)
	throws DOMException
    {
        throw new DOMException(DOMException.NOT_SUPPORTED_ERR,
			       "Method not supported");
    }

    public TypeInfo getSchemaTypeInfo() {
        throw new DOMException(DOMException.NOT_SUPPORTED_ERR,
			       "Method not supported");
    }

    public Object setUserData(String key, Object data, UserDataHandler handler)
    {
        throw new DOMException(DOMException.NOT_SUPPORTED_ERR,
			       "Method not supported");
    }

    public Object getUserData (String key)
    {
        throw new DOMException(DOMException.NOT_SUPPORTED_ERR,
			       "Method not supported");
    }
    public Object getFeature (String feature, String version)
    {
        throw new DOMException(DOMException.NOT_SUPPORTED_ERR,
			       "Method not supported");
    }

    public boolean isSameNode(Node node)
    {
        throw new DOMException(DOMException.NOT_SUPPORTED_ERR,
			       "Method not supported");
    }

    public boolean isEqualNode(Node node)
    {
        throw new DOMException(DOMException.NOT_SUPPORTED_ERR,
			       "Method not supported");
    }

    public String lookupNamespaceURI(String prefix)
    {
        throw new DOMException(DOMException.NOT_SUPPORTED_ERR,
			       "Method not supported");
    }

    public boolean isDefaultNamespace(String namespaceURI)
    {
        throw new DOMException(DOMException.NOT_SUPPORTED_ERR,
			       "Method not supported");
    }

    public String lookupPrefix(String namespaceURI)
    {
        throw new DOMException(DOMException.NOT_SUPPORTED_ERR,
			       "Method not supported");
    }

    public String getTextContent() throws DOMException
    {
        throw new DOMException(DOMException.NOT_SUPPORTED_ERR,
			       "Method not supported");
    }

    public void setTextContent(String textContent) throws DOMException
    {
        throw new DOMException(DOMException.NOT_SUPPORTED_ERR,
			       "Method not supported");
    }

    public short compareDocumentPosition(Node other) throws DOMException
    {
        throw new DOMException(DOMException.NOT_SUPPORTED_ERR,
			       "Method not supported");
    }

    public String getBaseURI()
    {
        throw new DOMException(DOMException.NOT_SUPPORTED_ERR,
			       "Method not supported");
    }
}
