/*
  COPYRIGHT NOTICE
  Copyright (c) 2012  Scott C. Neu and Arthur W. Toga

  This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package edu.ucla.loni.jdicom.plugin.write;

import java.io.IOException;
import javax.imageio.stream.ImageOutputStream;

/**
 * Block of encoded DICOM data in an output stream.
 *
 * @version 15 March 2011
 */
public class DicomBlock
{
    /** Output stream where the block is written. */
    private ImageOutputStream _outputStream;

    /** Offset in the stream to the start of the block. */
    private long _offset;

    /** Size of the block. */
    private long _size;

    /** True if the DICOM encoding is byte swapped; false if not. */
    private boolean _byteSwap;

    /**
     * Constructs a DICOM Block.
     *
     * @param outputStream Output stream where the block is written.
     * @param transferSyntax Transfer syntax that determines the encoding.
     *
     * @throws IOException If an I/O error occurs.
     */
    protected DicomBlock(ImageOutputStream outputStream, String transferSyntax)
	throws IOException
    {
	_outputStream = outputStream;
	_offset = _outputStream.getStreamPosition();

	// Byte swapping occurs for little endian byte ordering
	_byteSwap = !transferSyntax.equals("1.2.840.10008.1.2.2");
    }

    /**
     * Gets the stream offset.
     *
     * @return Offset in the stream to the start of the block of data.
     */
    public long getOffset()
    {
	return _offset;
    }

    /**
     * Gets the block size.
     *
     * @return Size of the block of data.
     */
    public long getSize()
    {
	return _size;
    }

    /**
     * Moves the block to another part of the output stream.
     *
     * @param newOffset New offset in the stream for the block.
     *
     * @throws IOException If an error occurs during reading or writing.
     */
    public void moveTo(long newOffset) throws IOException
    {
	// Move to the beginning of the block
	_outputStream.seek(_offset);

	// Read the block into memory
	byte[] contents = new byte[ (int)_size ];
	_outputStream.readFully(contents);

	// Move to the new offset
	_outputStream.seek(newOffset);

	// Write the block contents back to the stream
	_outputStream.write(contents);

	// Update the offset
	_offset = newOffset;
    }

    /**
     * Gets the output stream at the end of the block.
     *
     * @return Output stream at the end of the block.
     *
     * @throws IOException If an I/O error occurs.
     */
    protected ImageOutputStream _getOutputStreamAtEnd() throws IOException
    {
	// Move to the end of the block
	_outputStream.seek(_offset + _size);

	return _outputStream;
    }

    /**
     * Determines whether bytes are swapped.
     *
     * @return True if the DICOM encoding is byte swapped; false if not.
     */
    protected boolean _byteSwap()
    {
	return _byteSwap;
    }

    /**
     * Sets the stream offset of the block.
     *
     * @param offset Offset in the stream to the start of the block.
     */
    protected void _setOffset(long offset)
    {
	_offset = offset;
    }

    /**
     * Sets the block size between the offset and the current stream position.
     *
     * @throws IOException If an I/O error occurs.
     */
    protected void _setSize() throws IOException
    {
	_size = _outputStream.getStreamPosition() - _offset;
    }
}
