/*
  COPYRIGHT NOTICE
  Copyright (c) 2012  Scott C. Neu and Arthur W. Toga

  This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package edu.ucla.loni.jdicom.plugin.write;

import edu.ucla.loni.jdicom.DataElement;
import edu.ucla.loni.jdicom.DataSet;
import edu.ucla.loni.jdicom.encoder.ByteSwappedOutputStream;
import edu.ucla.loni.jdicom.encoder.DataElementEncoder;
import java.io.IOException;
import java.io.OutputStream;
import java.util.Enumeration;

/**
 * Output Stream implementation for the image metadata (DICOM tag 5200,9230).
 *
 * @version 16 March 2011
 */
public class ImageMetadataOS
{
    /** Output stream. */
    private ByteSwappedOutputStream _os;

    /** Transfer syntax that determines the encoding. */
    private String _transferSyntax;

    /** True if the DICOM encoding is byte swapped; false if not. */
    private boolean _byteSwap;

    /**
     * Constructs an Image Metadata OS.
     *
     * @param outputStream Output stream.
     * @param transferSyntax Transfer syntax that determines the encoding.
     *
     * @throws IOException If an error occurs during writing.
     */
    public ImageMetadataOS(OutputStream outputStream, String transferSyntax)
	throws IOException
    {
	_os = new ByteSwappedOutputStream(outputStream);
	_transferSyntax = transferSyntax;

	// Byte swapping occurs for little endian byte ordering
	_byteSwap = !transferSyntax.equals("1.2.840.10008.1.2.2");

	// Create the image metadata Data Element
	String tag = "52009230";
	String name = "Per-frame Functional Groups Sequence";
	String vr = "SQ";
	DataElement dataElement = new DataElement(tag, vr, name);

	// Write the start of the element
	DataElementEncoder encoder = new DataElementEncoder(_os,transferSyntax);
	encoder.writeDataElement(dataElement, false);

	// Write an undefined sequence length field
	_os.writeInt(0xFFFFFFFF, _byteSwap);
    }

    /**
     * Adds the data set to the end.
     *
     * @param dataSet Data Set containing the Data Elements to write.
     *
     * @throws IOException If an error occurs during writing.
     */
    public void append(DataSet dataSet) throws IOException
    {
	// Write the item tag
	_os.writeDataElementTag("FFFEE000", _byteSwap);

	// Write an undefined item length
	_os.writeInt(0xFFFFFFFF, _byteSwap);

	// Get an encoder
	DataElementEncoder encoder = new DataElementEncoder(_os,
							    _transferSyntax);
	// Write the data elements
	Enumeration enun = dataSet.getDataElements();
	while ( enun.hasMoreElements() ) {
	    encoder.writeDataElement( (DataElement)enun.nextElement() );
	}

	// Write the item delimitation tag
	_os.writeDataElementTag("FFFEE00D", _byteSwap);

	// Write a zero item length
	_os.writeInt(0, _byteSwap);
    }

    /**
     * Finishes writing the block.
     *
     * @throws IOException If an error occurs during writing.
     */
    public void finish() throws IOException
    {
	// Write the sequence delimitation item tag
	_os.writeDataElementTag("FFFEE0DD", _byteSwap);

	// Write a zero item length
	_os.writeInt(0, _byteSwap);
    }
}
