/*
  COPYRIGHT NOTICE
  Copyright (c) 2012  Scott C. Neu and Arthur W. Toga

  This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package edu.ucla.loni.nrrd.plugin;

import java.io.IOException;
import java.io.InputStream;
import java.util.zip.GZIPInputStream;
import javax.imageio.stream.ImageInputStream;
import javax.imageio.stream.ImageInputStreamImpl;

/**
 * Image Input Stream Impl that detects and reads compressed data in the GZIP
 * format.
 *
 * @version 28 October 2010
 */
public class GZipImageInputStream extends ImageInputStreamImpl
{
    /** Image Input Stream to read data from. */
    private ImageInputStream _imageStream;

    /** Original byte position of the Image Input Stream. */
    private long _imageStreamPosition;

    /** True if the data in the Image Input Stream is GZipped; false o/w. */
    private boolean _isGZipped;

    /** Input Stream that wraps around the Image Input Stream. */
    private InputStream _inputStream;

    /**
     * Constructs a GZIP Image Input Stream.
     *
     * @param imageStream Image Input Stream to read data from.
     *
     * @throws IOException If an I/O error occurs.
     * @throws IllegalStateException If the input stream is null.
     */
    public GZipImageInputStream(ImageInputStream imageStream) throws IOException
    {
	// No Image Input Stream
	if (imageStream == null) {
	    String msg = "No input has been set.";
	    throw new IllegalStateException(msg);
	}

	_imageStream = imageStream;
	_imageStreamPosition = imageStream.getStreamPosition();

	// Attempt to read a byte assuming GZIP compression
	try {
	    InputStream inputStream = new InputStreamAdapter(_imageStream);
	    GZIPInputStream gStream = new GZIPInputStream(inputStream);
	    gStream.read();

	    // Successful read
	    _isGZipped = true;
	}
	catch (Exception e) {}

	// Create the Input Stream
	_inputStream = _createInputStream();
    }

    /**
     * Reads a single byte from the stream.
     *
     * @return Value of the next byte in the stream, or -1 if EOF is reached.
     *
     * @throws IOException If an I/O error occurs.
     */
    public int read() throws IOException
    {
	// Make sure the stream is not closed
	checkClosed();

	// Reset the bit offset
	bitOffset = 0;

	// Read the byte
	int value = _inputStream.read();

	// Update the stream position
	if (value != -1) { streamPos++; }

	// Return the byte
	return value;
    }

    /**
     * Reads up to len bytes of data from the input stream into an array of
     * bytes.
     *
     * @param b The buffer into which the data is read.
     * @param off The start offset in array b at which the data is written.
     * @param len The maximum number of bytes to read.
     *
     * @return The total number of bytes read into the buffer, or -1 if there is
     *         no more data because the end of the stream has been reached.
     *
     * @throws IOException If an I/O error occurs.
     */
    public int read(byte b[], int off, int len) throws IOException
    {
	// Make sure the stream is not closed
	checkClosed();

	// Reset the bit offset
	bitOffset = 0;

	// Read bytes into the buffer
	int nBytes = _inputStream.read(b, off, len);

	// Update the stream position
	if (nBytes != -1) { streamPos += nBytes; }

	// Return the number of bytes read
	return nBytes;
    }

    /**
     * Sets the current stream position to the desired location.  The next read
     * will occur at this location.  The bit offset is set to 0.
     *
     * @param pos New file pointer position.
     *
     * @throws IOException If an I/O error occurs.
     */
    public void seek(long pos) throws IOException
    {
	// Make sure the stream is not closed
	checkClosed();

	// Reset the bit offset
	bitOffset = 0;

	// New position is before the current position
	if (pos < streamPos) {

	    // Reset the Input Stream
	    _inputStream = _createInputStream();

	    // Update the stream position
	    streamPos = 0;
	}

	// Nothing to do
	if (streamPos == pos) { return; }

	// New position is after the current position
	if (pos > streamPos) {

	    // Skip to the new position
	    long bytesToSkip = pos-streamPos;
	    long bytesSkipped = 0;
	    while (bytesSkipped < bytesToSkip) {
		bytesSkipped += _inputStream.skip(bytesToSkip-bytesSkipped);
	    }

	    // Update the stream position
	    streamPos = pos;
	}
    }

    /**
     * Closes the stream.
     *
     * @throws IOException If an I/O error occurs.
     */
    public void close() throws IOException
    {
	super.close();
	_inputStream.close();
    }

    /**
     * Called by the garbage collector when garbage collection determines that
     * there are no more references to the object.
     */
    protected void finalize() throws Throwable
    {
	// Do not let the GC close this stream
    }

    /**
     * Creates a new Input Stream that begins at the start of the Image Input
     * Stream.
     *
     * @return New Input Stream.
     *
     * @throws IOException If an I/O error occurs.
     */
    private InputStream _createInputStream() throws IOException
    {
	// Reset to the original position of the stream
	_imageStream.seek(_imageStreamPosition);

	// Create the Input Stream
	InputStream inputStream = new InputStreamAdapter(_imageStream);
	if (_isGZipped) { return new GZIPInputStream(inputStream); }
	return inputStream;
    }
}
