/*
  COPYRIGHT NOTICE
  Copyright (c) 2012  Scott C. Neu and Arthur W. Toga

  This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package edu.ucla.loni.nrrd.plugin;

import edu.ucla.loni.blobio.BlobReader;
import java.awt.image.DataBufferByte;
import java.io.IOException;
import javax.imageio.metadata.IIOMetadata;
import javax.imageio.stream.ImageInputStream;

/**
 * Blob Reader for the NRRD image format.
 *
 * @version 2 November 2010
 */
public class NrrdBlobReader extends BlobReader
{
    /** NRRD Image Reader used to decode the NRRD header. */
    private NrrdImageReader _imageReader;

    /** Stream metadata decoded from the input source. */
    private IIOMetadata _streamMetadata;

    /** Offset to the beginning of the blob data. */
    private long _beginBlobOffset;

    /** Offset to the end of the blob data. */
    private long _endBlobOffset;

    /**
     * Constructs an NRRD Blob Reader.
     *
     * @param origProvider The Blob Reader SPI that instantiated this Object.
     */
    public NrrdBlobReader(NrrdBlobReaderSpi origProvider)
    {
	super(origProvider);
    }

    /**
     * Sets the input source to the specified Object.  The input source must be
     * set before any of the query or read methods are used.  If the input is
     * null, any currently set input source will be removed.
     *
     * @param input Object which is the source of the input.
     *
     * @throws IllegalArgumentException If the input is not an instance of one
     *                                  of the classes returned by the
     *                                  originating service provider's
     *                                  getInputTypes method.
     */
    public void setInput(Object input)
    {
	// Create a new image reader
	_imageReader = new NrrdImageReader(new NrrdImageReaderSpi());

	// Set the input for the image reader
	_imageReader.setInput(input, false, false);

	// Get the input from the image reader
	super.setInput( _imageReader.getInput() );

	// Remove previously decoded metadata
	_streamMetadata = null;
    }

    /**
     * Returns the number of blobs available from the current input source.
     *
     * @param maxSize Maximum size of a blob.  A blob larger than the maximum
     *                size will be subdivided into blobs smaller than this size.
     *
     * @return Number of blobs available from the current input source.
     *
     * @throws IOException If an error occurs reading the information from the
     *                     input source.
     * @throws IllegalStateException If the input source has not been set.
     */
    public int getNumberOfBlobs(int maxSize) throws IOException
    {
	// Check the size
	if (maxSize <= 0) { return 0; }

	// Determine the blob size
	_getStreamMetadata();

	// Compute the number of blobs
	long blobSize = _endBlobOffset - _beginBlobOffset;
	return (int)Math.ceil(blobSize/maxSize);
    }

    /**
     * Returns an IIOMetadata object representing the metadata associated with
     * the input source as a whole (i.e., not associated with any particular
     * blob), or null if the reader does not support reading metadata, is
     * set to ignore metadata, or if no metadata is available.
     *
     * @return An IIOMetadata object, or null.
     *
     * @throws IOException If an error occurs during reading.
     * @throws IllegalStateException If the input source has not been set.
     */
    public IIOMetadata getStreamMetadata() throws IOException
    {
	return _getStreamMetadata();
    }

    /**
     * Returns an IIOMetadata object containing metadata associated with the
     * given blob, or null if no metadata is available.
     *
     * @param blobIndex Index of the blob to retrieve.
     * @param maxSize Maximum size of a blob.  A blob larger than the maximum
     *                size will be subdivided into blobs smaller than this size.
     *
     * @return An IIOMetadata object, or null.
     *
     * @throws IOException If an error occurs during reading.
     * @throws IllegalStateException If the input source has not been set.
     * @throws IndexOutOfBoundsException If the supplied index is out of bounds.
     */
    public IIOMetadata getBlobMetadata(int blobIndex, int maxSize)
	throws IOException
    {
	// Not applicable
	return null;
    }

    /**
     * Returns the specified binary large object (BLOB).
     *
     * @param blobIndex Index of the blob to retrieve.
     * @param maxSize Maximum size of a blob.  A blob larger than the maximum
     *                size will be subdivided into blobs smaller than this size.
     *
     * @return Binary object.
     *
     * @throws IOException If an error occurs during reading.
     * @throws IllegalStateException If the input source has not been set.
     * @throws IndexOutOfBoundsException If the supplied index is out of bounds.
     */
    public DataBufferByte getBlob(int blobIndex, int maxSize) throws IOException
    {
	// Check the size
	if (maxSize <= 0) { return new DataBufferByte(0); }

	// Get the total number of blobs
	int numOfBlobs = getNumberOfBlobs(maxSize);

	// Check the blob index
	if (blobIndex < 0 || blobIndex >= numOfBlobs) {
	    String msg = "Blob index of " + blobIndex + " is not in the " +
		"range (0," + (numOfBlobs-1) + ")";
	    throw new IndexOutOfBoundsException(msg);
	}

	// Move to the beginning of the blob
	ImageInputStream imageStream = _getPixelInputStream();
	imageStream.seek(_beginBlobOffset + blobIndex*maxSize);

	// Not last blob
	if (blobIndex < numOfBlobs-1) {
	    DataBufferByte blob = new DataBufferByte(maxSize);
	    imageStream.readFully( blob.getData() );
	    return blob;
	}

	// Last blob
	long blobSize = _endBlobOffset - imageStream.getStreamPosition();
	DataBufferByte blob = new DataBufferByte( (int)blobSize );
	imageStream.readFully( blob.getData() );
	return blob;
    }

    /**
     * Allows any resources held by this object to be released.  It is important
     * for applications to call this method when they know they will no longer
     * be using this Blob Reader.  Otherwise, the reader may continue to hold on
     * to resources indefinitely.
     */
    public void dispose()
    {
	// Attempt to close the image reader
	try { if (_imageReader != null) { _imageReader.dispose(); } }
	catch (Exception e) {}

	_imageReader = null;
    }

    /**
     * Gets the stream metadata.
     *
     * @return Stream metadata.
     *
     * @throws IOException If an error occurs during reading.
     * @throws IllegalStateException If the input source has not been set.
     */
    private IIOMetadata _getStreamMetadata() throws IOException
    {
	// Metadata has already been decoded
	if (_streamMetadata != null) { return _streamMetadata; }

	// Check the input
        if (_imageReader == null) {
	    throw new IllegalStateException("No input has been set.");
	}

	// Get the stream metadata
	_streamMetadata = _imageReader.getStreamMetadata();

	// Get the pixel image input stream
	ImageInputStream imageStream = _getPixelInputStream();

	// Determine the start and end of the blob data
	_beginBlobOffset = imageStream.getStreamPosition();
	_endBlobOffset = imageStream.length();

	// Return the stream metadata
	return _streamMetadata;
    }

    /**
     * Gets the pixel image input stream.
     *
     * @return Pixel image input stream.
     *
     * @throws IllegalStateException If the input source has not been set.
     */
    private ImageInputStream _getPixelInputStream()
    {
	// Check the input
        if (_input == null) {
	    throw new IllegalStateException("No input has been set.");
	}

	// Return the pixel image input stream
	if (_input instanceof NrrdInputStream) {
	    NrrdInputStream inputStream = (NrrdInputStream)_input;
	    return inputStream.getPixelDataInputStream();
	}
	else { return (ImageInputStream)_input; }
    }
}
