/*
  COPYRIGHT NOTICE
  Copyright (c) 2012  Scott C. Neu and Arthur W. Toga

  This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package edu.ucla.loni.nrrd.plugin;

import edu.ucla.loni.imageio.DualImageOutputStream;
import edu.ucla.loni.nrrd.NrrdHeader;
import edu.ucla.loni.nrrd.NrrdHeaderEncoder;
import java.awt.image.BufferedImage;
import java.awt.image.DataBuffer;
import java.awt.image.DataBufferByte;
import java.awt.image.DataBufferDouble;
import java.awt.image.DataBufferFloat;
import java.awt.image.DataBufferInt;
import java.awt.image.DataBufferShort;
import java.awt.image.DataBufferUShort;
import java.awt.image.Raster;
import java.awt.image.RenderedImage;
import java.io.IOException;
import java.io.OutputStream;
import java.util.Vector;
import java.util.zip.GZIPOutputStream;
import javax.imageio.IIOImage;
import javax.imageio.ImageTypeSpecifier;
import javax.imageio.ImageWriteParam;
import javax.imageio.ImageWriter;
import javax.imageio.metadata.IIOMetadata;
import javax.imageio.spi.ImageWriterSpi;
import javax.imageio.stream.ImageOutputStream;
import org.w3c.dom.Node;

/**
 * Image Writer for parsing and encoding NRRD images.
 *
 * @version 15 March 2012
 */
public class NrrdImageWriter extends ImageWriter
{
    /** NRRD Header of the image sequence currently being written. */
    private NrrdHeader _streamHeader;

    /** Number of images in the sequence currently being written. */
    private int _numberOfImagesWritten;

    /** GZipped output stream for pixel data. */
    private GZIPOutputStream _gZipOutputStream;

    /**
     * Constructs an NRRD Image Writer.
     *
     * @param originatingProvider The Image Writer Spi that instantiated this
     *                            Object.
     *
     * @throws IllegalArgumentException If the originating provider is not an
     *                                  NRRD Image Writer Spi.
     */
    public NrrdImageWriter(ImageWriterSpi originatingProvider)
    {
	super(originatingProvider);

	// Originating provider must be an NRRD Image Writer Spi
	if ( !(originatingProvider instanceof NrrdImageWriterSpi) ) {
	    String msg = "Originating provider must an NRRD Image Writer SPI.";
	    throw new IllegalArgumentException(msg);
	}
    }

    /**
     * Sets the destination to the given ImageOutputStream or other Object.  The
     * destination is assumed to be ready to accept data, and will not be closed
     * at the end of each write. If output is null, any currently set output
     * will be removed.
     *
     * @param output A Dual Image Output Stream or a Vector of two Image Output
     *               Streams to use for future writing.
     *
     * @throws IllegalArgumentException If output is not an instance of one of
     *                                  the classes returned by the originating
     *                                  service provider's getOutputTypes
     *                                  method.
     */
    public void setOutput(Object output)
    {
	// Vector of Image Output Streams
	if (output instanceof Vector) {
	    Vector streams = (Vector)output;

	    // Reset the output if one Image Output Stream
	    if (streams.size() == 1) { output = streams.firstElement(); }

	    // Otherwise convert to a Dual Image Output Stream
	    else { output = new DualImageOutputStream(streams); }
	}

	super.setOutput(output);

	_gZipOutputStream = null;
    }

    /**
     * Returns an IIOMetadata object containing default values for encoding a
     * stream of images.
     *
     * @param param An ImageWriteParam that will be used to encode the image, or
     *              null.
     *
     * @return An IIOMetadata object.
     */
    public IIOMetadata getDefaultStreamMetadata(ImageWriteParam param)
    {
	// Return a blank metadata object
	return new NrrdMetadata(null);
    }

    /**
     * Returns an IIOMetadata object containing default values for encoding an
     * image of the given type.
     *
     * @param imageType An ImageTypeSpecifier indicating the format of the image
     *                  to be written later.
     * @param param An ImageWriteParam that will be used to encode the image, or
     *              null.
     *
     * @return An IIOMetadata object.
     */
    public IIOMetadata getDefaultImageMetadata(ImageTypeSpecifier imageType,
					       ImageWriteParam param)
    {
	// Not applicable
	return null;
    }

    /**
     * Returns an IIOMetadata object that may be used for encoding and
     * optionally modified using its document interfaces or other interfaces
     * specific to the writer plug-in that will be used for encoding.
     *
     * @param inData An IIOMetadata object representing stream metadata, used to
     *               initialize the state of the returned object.
     * @param param An ImageWriteParam that will be used to encode the image, or
     *              null.
     *
     * @return An IIOMetadata object, or null if the plug-in does not provide
     *         metadata encoding capabilities.
     *
     * @throws IllegalArgumentException If inData is null.
     */
    public IIOMetadata convertStreamMetadata(IIOMetadata inData,
					     ImageWriteParam param)
    {
	// Only recognize NRRD Metadata
	if (inData instanceof NrrdMetadata) { return inData; }

	// Otherwise perform no conversion
	return null;
    }

    /**
     * Returns an IIOMetadata object that may be used for encoding and
     * optionally modified using its document interfaces or other interfaces
     * specific to the writer plug-in that will be used for encoding.
     *
     * @param inData An IIOMetadata object representing image metadata, used to
     *               initialize the state of the returned object.
     * @param imageType An ImageTypeSpecifier indicating the layout and color
     *                  information of the image with which the metadata will be
     *                  associated.
     * @param param An ImageWriteParam that will be used to encode the image, or
     *              null.
     *
     * @return An IIOMetadata object, or null if the plug-in does not provide
     *         metadata encoding capabilities.
     *
     * @throws IllegalArgumentException If either of inData or imageType is
     *                                  null.
     */
    public IIOMetadata convertImageMetadata(IIOMetadata inData,
					    ImageTypeSpecifier imageType,
					    ImageWriteParam param)
    {
	// Not applicable
	return null;
    }

    /**
     * Returns true if the methods that take an IIOImage parameter are capable
     * of dealing with a Raster (as opposed to RenderedImage) source image.
     *
     * @return True if Raster sources are supported.
     */
    public boolean canWriteRasters()
    {
	return true;
    }

    /**
     * Appends a complete image stream containing a single image and associated
     * stream and image metadata and thumbnails to the output.
     *
     * @param streamMetadata An IIOMetadata object representing stream metadata,
     *                       or null to use default values.
     * @param image An IIOImage object containing an image, thumbnails, and
     *              metadata to be written.
     * @param param An ImageWriteParam, or null to use a default
     *              ImageWriteParam.
     *
     * @throws IOException If an error occurs during writing.
     * @throws IllegalArgumentException If the image is null, if the stream
     *                                  metadata cannot be converted, or if the
     *                                  image type is unrecognized.
     * @throws IllegalStateException If the output has not been set.
     */
    public void write(IIOMetadata streamMetadata, IIOImage image,
		      ImageWriteParam param) throws IOException
    {
	// Write a sequence with 1 image
	prepareWriteSequence(streamMetadata);
	writeToSequence(image, param);
	endWriteSequence();
    }

    /**
     * Returns true if the writer is able to append an image to an image
     * stream that already contains header information and possibly prior
     * images.
     *
     * @return True If images may be appended sequentially.
     */
    public boolean canWriteSequence()
    {
	return true;
    }

    /**
     * Prepares a stream to accept a series of subsequent writeToSequence calls,
     * using the provided stream metadata object.  The metadata will be written
     * to the stream if it should precede the image data.
     *
     * @param streamMetadata A stream metadata object.
     *
     * @throws IOException If an error occurs during writing.
     * @throws IllegalArgumentException If the stream metadata is null or cannot
     *                                  be converted.
     * @throws IllegalStateException If the output has not been set.
     */
    public void prepareWriteSequence(IIOMetadata streamMetadata)
	throws IOException
    {
	// Convert the metadata to NRRD Metadata
	IIOMetadata iioMetadata = convertStreamMetadata(streamMetadata, null);

	// Unable to convert the metadata
	if (iioMetadata == null) {
	    String msg = "Unable to convert the stream metadata for encoding.";
	    throw new IllegalArgumentException(msg);
	}

	// Construct a stream NRRD Header from the metadata
	try {
	    String rootName = NrrdMetadataFormat.NATIVE_METADATA_FORMAT_NAME;
	    Node tree = iioMetadata.getAsTree(rootName);
	    _streamHeader = NrrdMetadataConversions.toNrrdHeader(tree);
	}
      
	catch (Exception e) {
	    String msg="Unable to construct an NRRD Header from the metadata.";
	    throw new IllegalArgumentException(msg);
	}

	// GZIP compression of non-domain first sample ordering not supported
	NrrdMetadata metadata = new NrrdMetadata(_streamHeader);
	if ( metadata.isGzipCompressed() && !metadata.varyDomainFirst() ) {
	    String msg = "Unable to write compressed images when \"kinds\" " +
		"starts with a non-domain dimension";
	    throw new IllegalArgumentException(msg);
	}

	// No output has been set
	if (output == null) {
	    String msg = "No output has been set.";
	    throw new IllegalStateException(msg);
	}

	// Get the image output stream for metadata
	boolean hasTwoStreams = false;
	ImageOutputStream metadataStream;
	if (output instanceof DualImageOutputStream) {
	    DualImageOutputStream dualStream = (DualImageOutputStream)output;
	    metadataStream = dualStream.getMetadataOutputStream();
	    hasTwoStreams = true;
	}
	else { metadataStream = (ImageOutputStream)output; }

	// Write the NRRD header to the metadata stream
	NrrdHeaderEncoder encoder = new NrrdHeaderEncoder();
	encoder.encode(_streamHeader, metadataStream, !hasTwoStreams);

	// Create a GZIP output stream if compression is specified
	if ( metadata.isGzipCompressed() ) {

	    // Get the image output stream for pixel data
	    ImageOutputStream pixelStream;
	    if (output instanceof DualImageOutputStream) {
		DualImageOutputStream dualStream=(DualImageOutputStream)output;
		pixelStream = dualStream.getPixelDataOutputStream();
	    }
	    else { pixelStream = (ImageOutputStream)output; }

	    // Create the GZIP output stream
	    OutputStream os = new OutputStreamAdapter(pixelStream);
	    _gZipOutputStream = new GZIPOutputStream(os);
	}

	_numberOfImagesWritten = 0;
    }

    /**
     * Appends a single image and possibly associated metadata and thumbnails,
     * to the output.  The supplied ImageReadParam is ignored.
     *
     * @param image An IIOImage object containing an image, thumbnails, and
     *              metadata to be written.
     * @param param An ImageWriteParam or null to use a default ImageWriteParam.
     *
     * @throws IOException If an error occurs during writing.
     * @throws IllegalArgumentException If the image is null or if the image
     *                                  type is not recognized.
     * @throws IllegalStateException If the output has not been set, or
     *                               prepareWriteSequence has not been called.
     */
    public void writeToSequence(IIOImage image, ImageWriteParam param)
	throws IOException
    { 
	// Check for the stream NRRD Header
	if (_streamHeader == null) {
	    String msg = "The image sequence has not been prepared.";
	    throw new IllegalStateException(msg);
	}

	// Check for a null image
	if (image == null) {
	    String msg = "Cannot write a null image.";
	    throw new IllegalArgumentException(msg);
	}

	// Get a Raster from the image
	Raster raster = image.getRaster();
	if (raster == null) {
	    RenderedImage renderedImage = image.getRenderedImage();

	    // If the Rendered Image is a Buffered Image, get it directly
	    if (renderedImage instanceof BufferedImage) {
		raster = ((BufferedImage)renderedImage).getRaster();
	    }

	    // Otherwise get a copy of the Raster from the Rendered Image
	    else { raster = renderedImage.getData(); }
	}

	// Update the Listeners
	processImageStarted(_numberOfImagesWritten);

	// Get the output stream for pixel data
	OutputStream pixelStream;
	if (_gZipOutputStream != null) {
	    pixelStream = _gZipOutputStream;
	}
	else if (output instanceof DualImageOutputStream) {
	    DualImageOutputStream dualStream = (DualImageOutputStream)output;
	    ImageOutputStream ios = dualStream.getPixelDataOutputStream();
	    pixelStream = new OutputStreamAdapter(ios);
	}
	else {
	    ImageOutputStream ios = (ImageOutputStream)output;
	    pixelStream = new OutputStreamAdapter(ios);
	}

	// Determine if the data is to be byte swapped
	NrrdMetadata metadata = new NrrdMetadata(_streamHeader);
	boolean byteSwap = metadata.isByteSwapped();
	DataBuffer dataBuffer = raster.getDataBuffer();

	// Color data
	if (metadata.hasColorImages() && dataBuffer instanceof DataBufferInt) {
	    int[] intArray = ((DataBufferInt)dataBuffer).getData();

	    // Separate the RGB components of each integer
	    byte[] byteArray = new byte[3*intArray.length];
	    for (int i = 0; i < intArray.length; i++) {
		int v = intArray[i];
		byteArray[3*i]   = (byte)((v >>> 16) & 0xFF);
		byteArray[3*i+1] = (byte)((v >>> 8) & 0xFF);
		byteArray[3*i+2] = (byte)((v >>> 0) & 0xFF);
	    }

	    // Write byte data
	    pixelStream.write(byteArray);
	}

	// Write byte data
	else if (dataBuffer instanceof DataBufferByte) {
	    pixelStream.write( ((DataBufferByte)dataBuffer).getData() );
	}

	// Write short data
	else if (dataBuffer instanceof DataBufferShort) {
	    short[] data = ((DataBufferShort)dataBuffer).getData();
	    _writeShorts(data, pixelStream, byteSwap);
	}

	// Write unsigned short data
	else if (dataBuffer instanceof DataBufferUShort) {
	    short[] data = ((DataBufferUShort)dataBuffer).getData();
	    _writeShorts(data, pixelStream, byteSwap);
	}

	// Write integer data
	else if (dataBuffer instanceof DataBufferInt) {
	    int[] data = ((DataBufferInt)dataBuffer).getData();
	    _writeInts(data, pixelStream, byteSwap);
	}

	// Write float data
	else if (dataBuffer instanceof DataBufferFloat) {
	    float[] data = ((DataBufferFloat)dataBuffer).getData();
	    _writeFloats(data, pixelStream, byteSwap);
	}

	// Write double data
	else if (dataBuffer instanceof DataBufferDouble) {
	    double[] data = ((DataBufferDouble)dataBuffer).getData();
	    _writeDoubles(data, pixelStream, byteSwap);
	}

	// Unrecognized type
	else {
	    String msg = "Unable to write the IIOImage.";
	    throw new IllegalArgumentException(msg);
	}

	// Update the Listeners
	_numberOfImagesWritten++;
	if ( abortRequested() ) { processWriteAborted(); }
	else { processImageComplete(); }
    }

    /**
     * Completes the writing of a sequence of images begun with
     * prepareWriteSequence.  Any stream metadata that should come at the end of
     * the sequence of images is written out, and any header information at the
     * beginning of the sequence is patched up if necessary.
     */
    public void endWriteSequence()
    {
	// Reset the stream NRRD Header
	_streamHeader = null;
    }

    /**
     * Allows any resources held by this object to be released.  It is important
     * for applications to call this method when they know they will no longer
     * be using this Image Writer.  Otherwise, the writer may continue to hold
     * on to resources indefinitely.
     */
    public void dispose()
    {
	// Attempt to close the output stream
	try {

	    // GZIP compression
	    if (_gZipOutputStream != null) {
		_gZipOutputStream.close();

		// Two Output Streams - close the metadata stream
		if (output instanceof DualImageOutputStream) {
		    DualImageOutputStream dualStream =
			(DualImageOutputStream)output;

		    dualStream.getMetadataOutputStream().close();
		}
	    }

	    // No compression
	    else {

		// Two Output Streams
		if (output instanceof DualImageOutputStream) {
		    ((DualImageOutputStream)output).close();
		}

		// One Output Stream
		else if (output instanceof ImageOutputStream) {
		    ((ImageOutputStream)output).close();
		}
	    }
	}
	catch (Exception e) {}

	output = null;
	_gZipOutputStream = null;
    }

    /**
     * Writes the shorts to the output stream.
     *
     * @param s Shorts to write.
     * @param outputStream Output stream to write to.
     * @param byteSwap True if bytes are to be swapped; false if not.
     *
     * @throws IOException If an error occurs during the writing.
     */
    private static void _writeShorts(short[] s, OutputStream outputStream,
				     boolean byteSwap)
	throws IOException
    {
	byte[] b = new byte[2*s.length];

	int boff = 0;
	if (byteSwap) {
	    for (int i = 0; i < s.length; i++) {
		b[boff++] = (byte)(s[i] >>> 0);
		b[boff++] = (byte)(s[i] >>> 8);
	    }
	}
	else {
	    for (int i = 0; i < s.length; i++) {
		b[boff++] = (byte)(s[i] >>> 8);
		b[boff++] = (byte)(s[i] >>> 0);
	    }
	}

	outputStream.write(b);
    }

    /**
     * Writes the integers to the output stream.
     *
     * @param i Integers to write.
     * @param outputStream Output stream to write to.
     * @param byteSwap True if bytes are to be swapped; false if not.
     *
     * @throws IOException If an error occurs during the writing.
     */
    private static void _writeInts(int[] i, OutputStream outputStream,
				   boolean byteSwap)
	throws IOException
    {
	byte[] b = new byte[4*i.length];

	int boff = 0;
	if (byteSwap) {
	    for (int j = 0; j < i.length; j++) {
		b[boff++] = (byte)(i[j] >>> 0);
		b[boff++] = (byte)(i[j] >>> 8);
		b[boff++] = (byte)(i[j] >>> 16);
		b[boff++] = (byte)(i[j] >>> 24);
	    }
	}
	else {
	    for (int j = 0; j < i.length; j++) {
		b[boff++] = (byte)(i[j] >>> 24);
		b[boff++] = (byte)(i[j] >>> 16);
		b[boff++] = (byte)(i[j] >>> 8);
		b[boff++] = (byte)(i[j] >>> 0);
	    }
	}

	outputStream.write(b);
    }

    /**
     * Writes the floats to the output stream.
     *
     * @param f Floats to write..
     * @param outputStream Output stream to write to.
     * @param byteSwap True if bytes are to be swapped; false if not.
     *
     * @throws IOException If an error occurs during the writing.
     */
    private static void _writeFloats(float[] f, OutputStream outputStream,
				     boolean byteSwap)
	throws IOException
    {
	byte[] b = new byte[4*f.length];

	int boff = 0;
	if (byteSwap) {
	    for (int i = 0; i < f.length; i++) {
		int v = Float.floatToIntBits(f[i]);
		b[boff++] = (byte)(v >>> 0);
		b[boff++] = (byte)(v >>> 8);
		b[boff++] = (byte)(v >>> 16);
		b[boff++] = (byte)(v >>> 24);
	    }
	}
	else {
	    for (int i = 0; i < f.length; i++) {
		int v = Float.floatToIntBits(f[i]);
		b[boff++] = (byte)(v >>> 24);
		b[boff++] = (byte)(v >>> 16);
		b[boff++] = (byte)(v >>> 8);
		b[boff++] = (byte)(v >>> 0);
	    }
	}

	outputStream.write(b);
    }

    /**
     * Writes the doubles to the output stream.
     *
     * @param d Doubles to write..
     * @param outputStream Output stream to write to.
     * @param byteSwap True if bytes are to be swapped; false if not.
     *
     * @throws IOException If an error occurs during the writing.
     */
    private static void _writeDoubles(double[] d, OutputStream outputStream,
				      boolean byteSwap)
	throws IOException
    {
        byte[] b = new byte[8*d.length];

        int boff = 0;
	if (byteSwap) {
	    for (int i = 0; i < d.length; i++) {
		long v = Double.doubleToLongBits(d[i]);
		b[boff++] = (byte)(v >>> 0);
		b[boff++] = (byte)(v >>> 8);
		b[boff++] = (byte)(v >>> 16);
		b[boff++] = (byte)(v >>> 24);
		b[boff++] = (byte)(v >>> 32);
		b[boff++] = (byte)(v >>> 40);
		b[boff++] = (byte)(v >>> 48);
		b[boff++] = (byte)(v >>> 56);
	    }
	}
	else {
	    for (int i = 0; i < d.length; i++) {
		long v = Double.doubleToLongBits(d[i]);
		b[boff++] = (byte)(v >>> 56);
		b[boff++] = (byte)(v >>> 48);
		b[boff++] = (byte)(v >>> 40);
		b[boff++] = (byte)(v >>> 32);
		b[boff++] = (byte)(v >>> 24);
		b[boff++] = (byte)(v >>> 16);
		b[boff++] = (byte)(v >>> 8);
		b[boff++] = (byte)(v >>> 0);
	    }
	}

	outputStream.write(b);
    }
}
