/*
  COPYRIGHT NOTICE
  Copyright (c) 2012  Scott C. Neu and Arthur W. Toga

  This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package edu.ucla.loni.nrrd.plugin;

import edu.ucla.loni.imageio.AppletFriendlyIIOMetadataFormat;
import java.util.Locale;
import java.util.MissingResourceException;
import java.util.ResourceBundle;
import javax.imageio.ImageTypeSpecifier;

/**
 * IIO Metadata Format that describes the tree structure of metadata in the
 * NRRD Metadata class.
 *
 * @version 25 October 2010
 */
public class NrrdMetadataFormat extends AppletFriendlyIIOMetadataFormat
{
    /** Name of the native metadata format. */
    public static final String NATIVE_METADATA_FORMAT_NAME =
	"edu_ucla_loni_nrrd_1.0";

    /** Name of the magic number attribute. */
    public static final String MAGIC = "magic";

    /** Name of the fields element. */
    public static final String FIELDS = "fields";

    /** Name of the key-value element. */
    public static final String KEYS = "keys";

    /** Name of the key attribute. */
    public static final String KEY = "key";

    /** Name of the value attribute. */
    public static final String VALUE = "value";

    /** Single instance of the NRRD Metadata Format. */
    private static NrrdMetadataFormat _format = new NrrdMetadataFormat();

    /** Constructs an NRRD Metadata Format. */
    private NrrdMetadataFormat()
    {
	// Create the root with some required children
	super(NATIVE_METADATA_FORMAT_NAME, CHILD_POLICY_SOME);

	// Magic number
	addAttribute(NATIVE_METADATA_FORMAT_NAME, MAGIC, DATATYPE_STRING, true,
		     null);

	// Field element with no children
	addElement(FIELDS, NATIVE_METADATA_FORMAT_NAME, CHILD_POLICY_EMPTY);

	// Keys element with no children
	addElement(KEYS, NATIVE_METADATA_FORMAT_NAME, CHILD_POLICY_EMPTY);
    }

    /**
     * Gets an instance of the NRRD Metadata Format.
     *
     * @return Single instance of the NRRD Metadata Format.
     */
    public static NrrdMetadataFormat getInstance()
    {
	return _format;
    }

    /**
     * Returns true if the element (and the subtree below it) is allowed to
     * appear in a metadata document for an image of the given type, defined by
     * an ImageTypeSpecifier.
     *
     * @param elementName The name of the element being queried.
     * @param imageType An ImageTypeSpecifier indicating the type of the image
     *                  that will be associated with the metadata.
     *
     * @return True if the node is meaningful for images of the given type.
     */
    public boolean canNodeAppear(String elementName,
				 ImageTypeSpecifier imageType)
    {
	// Not implemented
	return true;
    }

    /**
     * Returns a String containing a description of the named element, or null.
     *
     * @param elementName Name of the element.
     * @param locale Locale for which localization will be attempted.
     *
     * @return The element description, or null if it doesn't exist.
     *
     * @throws IllegalArgumentException If elementName is null, or is not a
     *                                  legal element name for this format.
     */
    public String getElementDescription(String elementName, Locale locale)
    {
	return _getDescription(elementName);
    }

    /**
     * Returns a String containing a description of the named attribute, or
     * null.
     *
     * @param elementName Name of the element.
     * @param attrName Name of the attribute.
     * @param locale Locale for which localization will be attempted, or null.
     *
     * @return The attribute description, or null if it doesn't exist.
     *
     * @throws IllegalArgumentException If elementName is null, or is not a
     *                                  legal element name for this format.
     * @throws IllegalArgumentException If attrName is null, or is not a legal
     *                                  legal attribute name for this element.
     */
    public String getAttributeDescription(String elementName, String attrName,
					  Locale locale)
    {
	// Any element is allowed to have these attributes
	if ( KEY.equals(attrName) || VALUE.equals(attrName) ) {
	    elementName = "*";
	}

	// Get the description
	String key = elementName + "/" + attrName;
	return _getDescription(key);
    }

    /**
     * Gets a description for the key.
     *
     * @param key Key to find a description for.
     *
     * @return Description for the key, or null if there isn't one.
     */
    private String _getDescription(String key)
    {
	// Get the description from the resource bundle
	try {
	    String name = getResourceBaseName();
	    Locale locale = Locale.getDefault();	
	    ResourceBundle bundle = ResourceBundle.getBundle(name, locale);

	    // Find the description
	    return bundle.getString(key);
	}

	// Return null if no resource bundle is available
	catch (MissingResourceException e) { return null; }
    }
}
